#!/usr/bin/env python
#
# __init__.py - FSLeyes - a python based OpenGL image viewer.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""*FSLeyes* - a 3D image viewer.

This package contains the application logic for *FSLeyes*.


.. image:: images/fsleyes.png
   :scale: 50%
   :align: center


--------
Overview
--------


*FSLeyes* is an OpenGL application for displaying 3D :mod:`overlays
<.overlay>`. All overlays are stored in a single list, the
:class:`.OverlayList`. Only one ``OverlayList`` ever exists - this list is
shared throughout the application.  The primary overlay type is the NIFTI
image format; and a range of other formats are also supported including MGH
volumes, GIFTI and Freesurfer surface files, and VTK triangle meshes.


Amongst other things, *FSLeyes* provides the following features:

  - Orthographic view  (:mod:`.orthopanel`)
  - Lightbox view (:mod:`.lightboxpanel`)
  - 3D view (:mod:`.scene3dpanel`)
  - Time series plotting (:mod:`.timeseriespanel`)
  - Histogram plotting (:mod:`.histogrampanel`)
  - Power spectrum plotting (:mod:`.powerspectrumpanel`)
  - Jupyter notebook integration (:mod:`.notebook`)
  - FSL atlas explorer (:mod:`.atlaspanel`)
  - FEAT cluster results explorer (:mod:`.clusterpanel`)
  - Melodic component classification (:mod:`.melodicclassificationpanel`)
  - NIFTI image editing (:mod:`.editor`)
  - A comprehensive command line interface (:mod:`.parseargs`)


*FSLeyes* makes heavy use of the :mod:`fsleyes_props` project, which is an
event-based programming framework.


------------
Entry points
------------


*FSLeyes* may be started with the :func:`fsleyes.main.main` function. *FSLeyes*
also includes an off-screen screenshot generator called `render`, which may
be started via the :func:`fsleyes.render.main` function.


---------------------------
Frames, views, and controls
---------------------------


The :class:`.FSLeyesFrame` is the top level GUI object. It is a container for
one or more *views*. All views are defined in the :mod:`.views` sub-package,
and are sub-classes of the :class:`.ViewPanel` class. Currently there are two
primary view categories - :class:`.CanvasPanel` views, which use :mod:`OpenGL`
to display overlays, and :class:`.PlotPanel` views, which use
:mod:`matplotlib` to plot data related to the overlays.


View panels may contain one or more *control* panels which provide an
interface allowing the user to control some aspect of the view (e.g. the
:class:`.OverlayDisplayToolBar`), or to display some other data associated
with the overlays (e.g. the :class:`.ClusterPanel`).  All controls are
sub-classes of the :class:`.ControlPanel` or :class:`.ControlToolBar` classes,
and all built-in controls are defined in the :mod:`.controls` sub-package.


The view/control panel class hierarchy is shown below:

.. graphviz::

   digraph hierarchy {

     graph [size=""];

     node [style="filled",
           fillcolor="#ffffdd",
           fontname="sans"];

     rankdir="BT";
     1  [label="panel.FSLeyesPanel"];
     2  [label="views.viewpanel.ViewPanel"];
     3  [label="controls.controlpanel.ControlPanel"];
     4  [label="views.plotpanel.PlotPanel"];
     5  [label="views.canvaspanel.CanvasPanel"];
     6  [label="views.histogrampanel.HistogramPanel"];
     7  [label="<other plot panels>"];
     8  [label="views.orthopanel.OrthoPanel"];
     9  [label="<other canvas panels>"];
     10 [label="controls.overlaylistpanel.OverlayListPanel"];
     11 [label="<other control panels>"];

     2  -> 1;
     3  -> 1;
     4  -> 2;
     5  -> 2;
     6  -> 4;
     7  -> 4;
     8  -> 5;
     9  -> 5;
     10 -> 3;
     11 -> 3;
   }

All toolbars inherit from the :class:`.FSLeyesToolBar` base class:

.. graphviz::

   digraph toolbar_hierarchy {

     graph [size=""];

     node [style="filled",
           fillcolor="#ffffdd",
           fontname="sans"];

     rankdir="BT";
     1 [label="toolbar.FSLeyesToolBar"];
     2 [label="controls.controlpanel.ControlToolBar"];
     3 [label="controls.overlaydisplaytoolbar.OverlayDisplayToolBar"];
     4 [label="controls.lightboxtoolbar.LightBoxToolBar"];
     5 [label="<other toolbars>"];

     2 -> 1;
     3 -> 2;
     4 -> 2;
     5 -> 2;
   }


----------------------
The ``DisplayContext``
----------------------


In order to manage how overlays are displayed, *FSLeyes* uses a
:class:`.DisplayContext`. Because *FSLeyes* allows multiple views to be opened
simultaneously, it needs to use multiple ``DisplayContext`` instances.
Therefore, one master ``DisplayContext`` instance is owned by the
:class:`FSLeyesFrame`, and a child ``DisplayContext`` is created for every
:class:`.ViewPanel`. The display settings managed by each child
``DisplayContext`` instance can be linked to those of the master instance;
this allows display properties to be synchronised across displays.


Each ``DisplayContext`` manages a collection of :class:`.Display` objects, one
for each overlay in the ``OverlayList``. Each of these ``Display`` objects
manages a single :class:`.DisplayOpts` instance, which contains overlay
type-specific display properties. Just as child ``DisplayContext`` instances
can be synchronised with the master ``DisplayContext``, child ``Display`` and
``DisplayOpts`` instances can be synchronised to the master instances.


The above description is summarised in the following diagram:


.. image:: images/fsleyes_architecture.png
   :scale: 40%
   :align: center


In this example, two view panels are open - an :class:`.OrthoPanel`, and a
:class:`.LightBoxPanel`. The ``DisplayContext`` for each of these views, along
with their ``Display`` and ``DisplayOpts`` instances (one of each for every
overlay in the ``OverlayList``) are linked to the master ``DisplayContext``
(and its ``Display`` and ``DisplayOpts`` instances), which is managed by the
``FSLeyesFrame``.  All of this synchronisation functionality is provided by
the ``props`` package.


See the :mod:`~fsleyes.displaycontext` package documentation for more
details.


-----------------------
Events and notification
-----------------------

TODO


.. note:: The current version of FSLeyes (|version|) lives in the
          :mod:`fsleyes.version` module.
"""


import os.path as op
import            os
import            sys
import            logging
import            warnings

import fsl.utils.settings as fslsettings
import fsleyes_widgets    as fwidgets
import fsleyes.version    as version


# The logger is assigned in
# the configLogging function
log = None


# If set to True, logging will not be configured
disableLogging = True


__version__ = version.__version__
"""The current *FSLeyes* version (read from the :mod:`fsleyes.version`
module).
"""


assetDir = op.abspath(op.join(op.dirname(__file__), 'assets'))
"""Directory which contains all *FSLeyes* assets/resources (e.g. icon
files).
"""


def canWriteToAssetDir():
    """Returns ``True`` if the user can write to the FSLeyes asset directory,
    ``False`` otherwise.
    """
    return os.access(assetDir, os.W_OK | os.X_OK)


def initialise():
    """Called when `FSLeyes`` is started as a standalone application.  This
    function *must* be called before most other things in *FSLeyes* are used.
    """
    import fsleyes.plugins as plugins

    # implement various hacks and workarounds
    _hacksAndWorkarounds()

    # Initialise the fsl.utils.settings module
    fslsettings.initialise('fsleyes')

    # initialise FSLeyes plugins (will discover
    # any plugins saved in the settings dir)
    plugins.initialise()

    # The fsleyes.actions.frameactions module
    # monkey-patches some things into the
    # FSLeyesFrame class, so it must be
    # imported immediately after fsleyes.frame.
    import fsleyes.frame                 # noqa
    import fsleyes.actions.frameactions  # noqa


def _hacksAndWorkarounds():
    """Called by :func:`initialise`. Implements hacks and workarounds for
    various things.
    """

    # Under wxPython/Phoenix, the
    # wx.html package must be imported
    # before a wx.App has been created
    import wx.html  # noqa

    # nibabel rejects NIfTI images where the
    # quaternion vector has a length greater
    # than 1. This is fine, as it is mandated
    # by the NIfTI spec. But FSL is much more
    # lenient than nibabel, and nibabel can
    # also reject some qforms due to float32
    # imprecision. So here we're increasing
    # the tolerance of nibabel to strange
    # qforms.
    import nibabel as nib
    nib.Nifti1Header.quaternion_threshold = -1e5

    # OSX sometimes sets the local environment
    # variables to non-standard values, which
    # breaks the python locale module.
    #
    # http://bugs.python.org/issue18378
    try:
        import locale
        locale.getdefaultlocale()
    except ValueError:
        os.environ['LC_ALL'] = 'C.UTF-8'


def configLogging(verbose=0, noisy=None):
    """Configures *FSLeyes* ``logging``.

    :arg verbose: A number between 0 and 3, indicating the verbosity level.
    :arg noisy:   A sequence of module names - logging will be enabled on these
                  modules.
    """

    global log

    # already configured
    if log is not None:
        return

    if noisy is None:
        noisy = []

    # Set up the root logger
    logFormatter = logging.Formatter('%(levelname)8.8s '
                                     '%(filename)20.20s '
                                     '%(lineno)4d: '
                                     '%(funcName)-15.15s - '
                                     '%(message)s')
    logHandler  = logging.StreamHandler()
    logHandler.setFormatter(logFormatter)

    log = logging.getLogger()
    log.addHandler(logHandler)

    # Everything below this point sets up verbosity
    # as requested by the user. But verbosity-related
    # command line arguments are not exposed to the
    # user if disableLogging is set.
    if disableLogging:
        return

    # show dep warnings
    warnings.filterwarnings('default', category=DeprecationWarning)

    # Now we can set up logging
    if verbose == 1:
        log.setLevel(logging.DEBUG)

        # make some noisy things quiet
        logging.getLogger('fsleyes.gl')     .setLevel(logging.WARNING)
        logging.getLogger('fsleyes.views')  .setLevel(logging.WARNING)
        logging.getLogger('fsleyes_props')  .setLevel(logging.WARNING)
        logging.getLogger('fsleyes_widgets').setLevel(logging.WARNING)
    elif verbose == 2:
        log.setLevel(logging.DEBUG)
        logging.getLogger('fsleyes_props')  .setLevel(logging.WARNING)
        logging.getLogger('fsleyes_widgets').setLevel(logging.WARNING)
    elif verbose == 3:
        log.setLevel(logging.DEBUG)
        logging.getLogger('fsleyes_props')  .setLevel(logging.DEBUG)
        logging.getLogger('fsleyes_widgets').setLevel(logging.DEBUG)

    for mod in noisy:
        logging.getLogger(mod).setLevel(logging.DEBUG)

    # The trace module monkey-patches some
    # things if its logging level has been
    # set to DEBUG, so we import it now so
    # it can set itself up.
    traceLogger = logging.getLogger('fsleyes_props.trace')
    if traceLogger.getEffectiveLevel() <= logging.DEBUG:
        import fsleyes_props.trace  # noqa


# Just for convenience
from fsleyes.main import embed, shutdown  # noqa



#!/usr/bin/env python
#
# clusterpanel.py - The ClusterPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ClusterPanel` class, a *FSLeyes control*
panel for viewing cluster results from a FEAT analysis.
"""

import                                  logging
import                                  wx
import                                  six

import fsl.utils.idle                as idle
import fsl.data.image                as fslimage
import fsl.data.featimage            as featimage
import fsl.data.featanalysis         as featanalysis

import fsleyes_widgets               as fwidgets
import fsleyes_widgets.utils.status  as status
import fsleyes_widgets.widgetgrid    as widgetgrid
import fsleyes.views.canvaspanel     as canvaspanel
import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.strings               as strings
import fsleyes.autodisplay           as autodisplay


log = logging.getLogger(__name__)


class ClusterPanel(ctrlpanel.ControlPanel):
    """The ``ClusterPanel`` shows a table of cluster results from the analysis
    associated with a :class:`.FEATImage` overlay. A ``ClusterPanel`` looks
    something like the following:

    .. image:: images/clusterpanel.png
       :scale: 50%
       :align: center

    The ``ClusterPanel`` contains controls which allow the user to:

      - Select the COPE for which cluster results are displayed

      - Add a Z statistic overlay for the currently displayed COPE

      - Add a cluster mask overlay for the currently displayed COPE

      - Navigate to the Z maximum location, Z centre-of-gravity location,
        or COPE maximum location, for a specific cluster.


    When an overlay is selected (detected via the
    :attr:`.DisplayContext.selectedOverlay` property), a ``ClusterPanel``
    tries to identify the FEAT analysis associated with the overlay. If
    the overlay is a :class:`.FEATImage`, then no work needs to be done.
    Otherwise, the ``ClusterPanel`` uses functions in the :mod:`.featanalysis`
    module to identify the FEAT directory. Ultimately, a reference to a
    :class:`.FEATImage` associated with the currently selected overlay is
    obtained, so the ``ClusterPanel`` can retrieve contrast and cluster
    information.


    A ``ClusterPanel`` uses a :class:`.WidgetGrid` to display information
    about the clusters associated with a contrast. Because creating all of the
    widgets contained in the ``WidgetGrid`` is expensive, a ``ClusterPanel``
    creates (on demand) and caches ``WidgetGrid`` instances for all
    :class:`.FEATImage` overlay and contrasts. This means that when the user
    changes the currently selected overlay, or the current contrast, the
    displayed cluster information is updated quickly.
    """


    @staticmethod
    def supportedViews():
        """The :class:`.ClusterPanel` is restricted for use with
        ``OrthoPanel``, ``LightBoxPanel`` and ``Scene3DPanel`` views.
        """
        return [canvaspanel.CanvasPanel]


    @staticmethod
    def defaultLayout():
        """Returns a dictionary containing layout settings to be passed to
        :class:`.ViewPanel.togglePanel`.
        """
        return {'location' : wx.TOP}


    def __init__(self, parent, overlayList, displayCtx, viewPanel):
        """Create a ``ClusterPanel``.

        :arg parent:      The :mod:`wx` parent object.

        :arg overlayList: The :class:`.OverlayList` instance.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg viewPanel:   The :class:`.ViewPanel` instance.
        """
        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, viewPanel)

        self.__disabledText = wx.StaticText(
            self,
            style=(wx.ALIGN_CENTRE_HORIZONTAL |
                   wx.ALIGN_CENTRE_VERTICAL))

        self.__overlayName  = wx.StaticText(self)
        self.__addZStats    = wx.Button(    self)
        self.__addClustMask = wx.Button(    self)
        self.__statSelect   = wx.Choice(    self)

        # The featImages dictionary is a mapping
        # of { overlay : FEATImage } pairs. The
        # ClusterPanel will work with any overlay,
        # if that overlay is contained within a
        # feat analysis. The selectedOverlayChanged
        # method identifies the FEAT analysis
        # associated with the newly selected overlay,
        # and stores the mapping here. Note that the
        # FEATImage associated with an overlay might
        # not be in the overlay list.
        self.__featImages = {}

        # A WidgetGrid is created for each
        # contrast of a FEAT image, and cached
        # in this dictionary. This is because
        # it is quite expensive to create the
        # grid widgets. This dictionary contains
        # {FEATImage : [WidgetGrid]} mappings.
        self.__clusterGrids = {}

        self.__addZStats   .SetLabel(strings.labels[self, 'addZStats'])
        self.__addClustMask.SetLabel(strings.labels[self, 'addClustMask'])

        self.__sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.__sizer)

        self.__topSizer  = wx.BoxSizer(wx.HORIZONTAL)
        self.__mainSizer = wx.BoxSizer(wx.VERTICAL)

        args = {'flag' : wx.EXPAND, 'proportion' : 1}

        self.__topSizer.Add(self.__statSelect,   flag=wx.EXPAND, proportion=1)
        self.__topSizer.Add(self.__addZStats,    flag=wx.EXPAND, proportion=1)
        self.__topSizer.Add(self.__addClustMask, flag=wx.EXPAND, proportion=1)

        self.__mainSizer.Add(self.__overlayName, flag=wx.EXPAND)
        self.__mainSizer.Add(self.__topSizer,    flag=wx.EXPAND)

        # Only one of the disabledText or
        # mainSizer are shown at any one time
        self.__sizer.Add(self.__disabledText, **args)
        self.__sizer.Add(self.__mainSizer,    **args)

        overlayList.addListener('overlays',
                                self.name,
                                self.__overlayListChanged)
        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__selectedOverlayChanged)

        self.__statSelect  .Bind(wx.EVT_CHOICE, self.__statSelected)
        self.__addZStats   .Bind(wx.EVT_BUTTON, self.__addZStatsClick)
        self.__addClustMask.Bind(wx.EVT_BUTTON, self.__addClustMaskClick)

        self.SetMinSize(self.__calcMinSize())

        self.__selectedOverlay = None
        self.__selectedOverlayChanged()


    def destroy(self):
        """Must be called when this ``ClusterPanel`` is no longer needed.
        Removes some property listeners, and calls
        :meth:`.ControlPanel.destroy`.
        """

        for grids in self.__clusterGrids.values():
            for grid in grids:
                if grid is not None:
                    grid.Destroy()

        self.__clusterGrids = None
        self.__featImages   = None

        self.overlayList.removeListener('overlays',        self.name)
        self.displayCtx .removeListener('selectedOverlay', self.name)

        ctrlpanel.ControlPanel.destroy(self)


    def __calcMinSize(self):
        """Figures out the minimum size that this ``ClusterPanel`` should
        have. Called by :meth:`__init__`.

        When the ``ClusterPanel`` is created, the COPE combo box is not
        populated, so has no minimum size. Here, we figure out a good minimum
        size for it. We can then calculate a good minimum size for the entire
        panel.
        """

        dc = wx.ClientDC(self.__statSelect)

        dummyName = strings.labels[self, 'clustName'].format(1, 'WW')

        w, h = dc.GetTextExtent(dummyName)

        self.__statSelect .SetMinSize((w, h))
        self.__sizer.Layout()

        return self.__sizer.GetMinSize()


    def __disable(self, message):
        """Disables the ``ClusterPanel``, and displays the given message.
        Called when the selected overlay is not associated with a FEAT
        analysis, or when cluster results cannot be displayed for some reason.
        """

        self.__disabledText.SetLabel(message)
        self.__sizer.Show(self.__disabledText, True)
        self.__sizer.Show(self.__mainSizer,    False)
        self.Layout()


    def __enable(self):
        """Enables the ``ClusterPanel``. This amounts to hiding the
        disabled message panel, and showing the appropriate
        :class:`.WidgetGrid` for the currently selected overlay / contrast.
        The appropriate ``WidgetGrid`` is assumed to have already been
        created.
        """
        self.__sizer.Show(self.__disabledText, False)
        self.__sizer.Show(self.__mainSizer,    True)

        overlay   = self.__selectedOverlay
        featImage = self.__featImages[overlay]
        contrast  = self.__statSelect.GetSelection()

        for fimg, grids in self.__clusterGrids.items():
            for i, grid in enumerate(grids):
                if grid is not None:
                    show = fimg is featImage and i == contrast
                    self.__mainSizer.Show(grid, show)

        self.Layout()


    def __statSelected(self, ev=None):
        """Called when a COPE is selected. Retrieves a cached
        :class:`.WidgetGrid`, or creates a new one (via the
        :meth:`__genClusterGrid` method) which displays information
        about the clusters associated with the currently selected contrast.
        """
        overlay   = self.__selectedOverlay
        featImage = self.__featImages[overlay]
        idx       = self.__statSelect.GetSelection()
        data      = self.__statSelect.GetClientData(idx)

        self.Refresh()
        self.Update()

        if featImage not in self.__clusterGrids:
            self.__clusterGrids[featImage] = [None] * featImage.numContrasts()

        grid = self.__clusterGrids[featImage][idx]

        if grid is None:
            grid = self.__genClusterGrid(overlay, featImage, idx, data)
            self.__clusterGrids[featImage][idx] = grid
            self.__mainSizer.Add(grid, flag=wx.EXPAND, proportion=1)

        self.__enable()
        self.__enableOverlayButtons()


    def __addZStatsClick(self, ev):
        """Called when the *Add Z statistics* button is pushed. Retrieves
        the Z statistic image for the current COPE (see the
        :meth:`.FEATImage.getZStats` method), and adds it as an overlay
        to the :class:`.OverlayList`.
        """

        overlay   = self.__selectedOverlay
        featImage = self.__featImages[overlay]
        contrast  = self.__statSelect.GetSelection()
        zstats    = featImage.getZStats(contrast)

        for ol in self.overlayList:

            # Already in overlay list
            if ol.dataSource == zstats.dataSource:
                return

        log.debug('Adding Z-statistic {} to overlay list'.format(zstats.name))
        self.overlayList.append(zstats, overlayType='volume')

        zthres = featImage.thresholds()['z']

        autodisplay.autoDisplay(
            zstats,
            self.overlayList,
            self.displayCtx,
            zthres=zthres,
            posCmap='red-yellow',
            negCmap='blue-lightblue')


    def __addClustMaskClick(self, ev):
        """Called when the *Add cluster mask* button is pushed. Retrieves the
        cluster mask image for the currewnt contrast (see the
        :meth:`.FEATImage.getClusterMask` method)
        """
        overlay   = self.__selectedOverlay
        featImage = self.__featImages[overlay]
        contrast  = self.__statSelect.GetSelection()
        mask      = featImage.getClusterMask(contrast)

        for ol in self.overlayList:

            # Already in overlay list
            if ol.dataSource == mask.dataSource:
                return

        log.debug('Adding cluster mask {} to overlay list'.format(mask.name))
        self.overlayList.append(mask, overlayType='label')


    def __genClusterGrid(self, overlay, featImage, contrast, clusters):
        """Creates and returns a :class:`.WidgetGrid` which contains the given
        list of clusters, which are related to the given contrast.


        .. note:: This method assumes that the given ``overlay`` is an
                  :class:`.Image` which has the same voxel dimensions as,
                  and shares the the same world coordinate system as the
                  ``featImage``.


        :arg overlay:   The overlay for which clusters are currently being
                        displayed.

        :arg featImage: The :class:`.FEATImage` to which the clusters are
                        related.

        :arg contrast:  The (0-indexed) number of the contrast to which the
                        clusters are related.

        :arg clusters:  A sequence of objects, each representing one cluster.
                        See the :meth:`.FEATImage.clusterResults` method.
        """

        cols = {'index'         : 0,
                'nvoxels'       : 1,
                'p'             : 2,
                'logp'          : 3,
                'zmax'          : 4,
                'zmaxcoords'    : 5,
                'zcogcoords'    : 6,
                'copemax'       : 7,
                'copemaxcoords' : 8,
                'copemean'      : 9}

        grid    = widgetgrid.WidgetGrid(self)
        conName = featImage.contrastNames()[contrast]
        opts    = self.displayCtx.getOpts(overlay)

        # We hide the grid and disable
        # this panle while the grid is
        # being created.
        grid.Hide()
        self.Disable()

        grid.SetGridSize(len(clusters), 10)

        grid.ShowRowLabels(False)
        grid.ShowColLabels(True)

        for col, i in cols.items():
            grid.SetColLabel(i, strings.labels[self, col])

        def makeCoordButton(coords):

            label = wx.StaticText(grid, label='[{} {} {}]'.format(*coords))
            btn   = wx.Button(grid,
                              label=six.u('\u2192'),
                              style=wx.BU_EXACTFIT)

            sizer = wx.BoxSizer(wx.HORIZONTAL)
            sizer.Add(label, flag=wx.EXPAND, proportion=1)
            sizer.Add(btn)

            def onClick(ev):
                dloc = opts.transformCoords([coords], 'voxel', 'display')[0]
                self.displayCtx.location = dloc

            btn.Bind(wx.EVT_BUTTON, onClick)

            return sizer

        # Creating all of the widgets could
        # take a bit of time, so we'll
        # do it asynchronously via idle.idle
        # display a message while doing so.
        status.update(strings.messages[self, 'loadingCluster'].format(
            contrast + 1, conName), timeout=None)

        def addCluster(i, clust):

            if not fwidgets.isalive(grid):
                return

            zmaxbtn    = makeCoordButton((clust.zmaxx,
                                          clust.zmaxy,
                                          clust.zmaxz))
            zcogbtn    = makeCoordButton((clust.zcogx,
                                          clust.zcogy,
                                          clust.zcogz))
            copemaxbtn = makeCoordButton((clust.copemaxx,
                                          clust.copemaxy,
                                          clust.copemaxz))

            def fmt(v):
                return '{}'.format(v)

            grid.SetText(  i, cols['index'],         fmt(clust.index))
            grid.SetText(  i, cols['nvoxels'],       fmt(clust.nvoxels))
            grid.SetText(  i, cols['p'],             fmt(clust.p))
            grid.SetText(  i, cols['logp'],          fmt(clust.logp))
            grid.SetText(  i, cols['zmax'],          fmt(clust.zmax))
            grid.SetWidget(i, cols['zmaxcoords'],    zmaxbtn)
            grid.SetWidget(i, cols['zcogcoords'],    zcogbtn)
            grid.SetText(  i, cols['copemax'],       fmt(clust.copemax))
            grid.SetWidget(i, cols['copemaxcoords'], copemaxbtn)
            grid.SetText(  i, cols['copemean'],      fmt(clust.copemean))

        # Refresh the grid widget when all
        # clusters have been added.
        def onFinish():

            if not fwidgets.isalive(grid):
                return

            status.update('All clusters loaded.')
            self.Enable()
            grid.Show()
            grid.Refresh()

        for i, clust in enumerate(clusters):
            idle.idle(addCluster, i, clust)

        idle.idle(onFinish)

        return grid


    def __overlayListChanged(self, *a):
        """Called when the :class:`.OverlayList` changes. Destroys any
        cached :class:`.WidgetGrid` instances as needed, updates the *Add Z
        statistic* and *Add cluster mask* buttons, in case the user removed
        them, and calls :meth:`__selectedOverlayChanged`.
        """

        # Remove and destroy any cached
        # WidgetGrid panels for overlays
        # that have been removed from the
        # list.
        for overlay in list(self.__featImages.keys()):
            if overlay not in self.overlayList:

                featImage = self.__featImages.pop(overlay)

                # Has the feat image associated with
                # this overlay also been removed?
                if featImage is overlay or \
                   featImage not in self.overlayList:

                    # The grid widgets for the feat image
                    # associated with this overlay may
                    # have already been destroyed.
                    try:             grids = self.__clusterGrids.pop(featImage)
                    except KeyError: grids = []

                    for grid in grids:
                        if grid is not None:
                            self.__mainSizer.Detach(grid)
                            grid.Destroy()

        self.__selectedOverlayChanged()
        self.__enableOverlayButtons()


    def __enableOverlayButtons(self):
        """Enables/disables the *Add Z statistic* and *Add cluster mask*
        buttons depending on whether the corresponding overlays are in the
        :class:`.OverlayList`.
        """

        if self.__selectedOverlay is None:
            return

        overlay   = self.__selectedOverlay
        featImage = self.__featImages[overlay]
        contrast  = self.__statSelect.GetSelection()

        # No cluster results
        if contrast < 0:
            self.__addZStats   .Enable(False)
            self.__addClustMask.Enable(False)
            return

        zstat     = featImage.getZStats(     contrast)
        clustMask = featImage.getClusterMask(contrast)

        dss = [ovl.dataSource for ovl in self.overlayList]

        self.__addZStats   .Enable(zstat    .dataSource not in dss)
        self.__addClustMask.Enable(clustMask.dataSource not in dss)


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` changes,
        and by the :meth:`__overlayListChanged` method.

        If the newly selected overlay is a :class:`.FEATImage` (or is otherwise
        associated with a FEAT analysis) which has cluster results, they are
        loaded in, and displayed on a :class:`.WidgetGrid`.
        """

        prevOverlay            = self.__selectedOverlay
        self.__selectedOverlay = None

        # No overlays are loaded
        if len(self.overlayList) == 0:
            self.__disable(strings.messages[self, 'noOverlays'])
            return

        overlay = self.displayCtx.getSelectedOverlay()

        # Overlay is in-memory
        if overlay.dataSource is None:
            self.__disable(strings.messages[self, 'notFEAT'])
            return

        featDir = featanalysis.getAnalysisDir(overlay.dataSource)

        # No FEAT analysis, or not an Image,
        # can't do anything with that
        if featDir is None or not isinstance(overlay, fslimage.Nifti):
            log.debug('Overlay {} is not part of a feat '
                      'analysis, or is not Nifti'.format(overlay))
            self.__disable(strings.messages[self, 'notFEAT'])
            return

        # Selected overlay is either the
        # same one (maybe the overlay list,
        # rather than the selected overlay,
        # changed) or the newly selected
        # overlay is from the same FEAT
        # analysis. No need to do anything.
        if prevOverlay is not None:

            prevFeatImage = self.__featImages.get(prevOverlay)

            if prevOverlay is overlay or \
               (prevFeatImage is not None and
                    featDir == prevFeatImage.getFEATDir()):

                log.debug('Overlay {} is already selected.'.format(overlay))

                # Make sure the overlay -> FEATImage
                # mapping is present, and (re-)cache
                # a reference to the selected overlay.
                self.__featImages[overlay] = prevFeatImage
                self.__selectedOverlay     = overlay

                return

        # We're in business. The newly selected
        # overlay is a part of a FEAT analysis
        # which is not currently being displayed.
        self.__selectedOverlay = overlay

        # Clear the stat selection combo box.
        self.__statSelect.Clear()

        # Get the FEATImage associated with
        # this overlay, so we can get
        # information about the FEAT analysis
        featImage = self.__featImages.get(overlay)
        if featImage is None:

            # If the overlay itself is a FEATImage,
            # then we have nothing to do.
            if isinstance(overlay, featimage.FEATImage):
                featImage = overlay
            else:
                # The FEATImage might already
                # be in the overlay list -
                # let's search for it.
                for ovl in self.overlayList:
                    if isinstance(ovl, featimage.FEATImage) and \
                       ovl.getFEATDir() == featDir:
                        featImage = ovl

            # As a last resort, if the FEATImage is not
            # in the overlay list, we'll create one.
            if featImage is None:
                featImage = featimage.FEATImage(featDir,
                                                loadData=False,
                                                calcRange=False)

            self.__featImages[overlay] = featImage

        log.debug('Identified FEAT analysis associated with overlay '
                  '{}: {}'.format(overlay, featImage.getFEATDir()))

        # Get the contrast and cluster
        # information for the FEAT analysis.
        display  = self.displayCtx.getDisplay(overlay)
        numCons  = featImage.numContrasts()
        conNames = featImage.contrastNames()

        try:
            # clusts is a list of (contrast, clusterList) tuples
            clusts = [(c, featImage.clusterResults(c)) for c in range(numCons)]
            clusts = [c for c in clusts if c[1] is not None]

        # Error parsing the cluster data
        except Exception as e:
            log.warning('Error parsing cluster data for '
                        '{}: {}'.format(featImage.name, str(e)), exc_info=True)
            self.__disable(strings.messages[self, 'badData'])
            return

        # No cluster results exist
        # for any contrast
        if len(clusts) == 0:
            self.__disable(strings.messages[self, 'noClusters'])
            return

        # Populate the stat selection combo box
        for contrast, clusterList in clusts:
            name = conNames[contrast]
            name = strings.labels[self, 'clustName'].format(contrast + 1, name)

            self.__statSelect.Append(name, clusterList)

        self.__overlayName.SetLabel(display.name)

        # Refresh the widget grid
        self.__statSelect.SetSelection(0)
        self.__statSelected()

        self.Layout()



#!/usr/bin/env python
#
# offscreenscene3dcanvas.py - The OffScreenScene3DCanvas.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OffScreenScene3DCanvas`, which is a
:class:`.Scene3DCanvas` for use in off-screen OpenGL rendering.
"""


import fsleyes.gl           as fslgl
from . import scene3dcanvas as sc


class OffScreenScene3DCanvas(sc.Scene3DCanvas, fslgl.OffScreenCanvasTarget):
    """The ``OffScreensScene3DCanvas`` is a :class:`.Scene3DCanvas` which uses
    a :class:`.RenderTexture` as its target, for static off-screen OpenGL
    rendering.
    """

    def __init__(self,
                 overlayList,
                 displayCtx,
                 *args,
                 **kwargs):
        """Create an ``OffScreenScene3DCanvas``.

        See the :class:`.Scene3DCanvas` and :class:`.OffscreenCanvasTarget`
        classes for details on the other arguments.
        """

        fslgl.OffScreenCanvasTarget.__init__(self, *args, **kwargs)
        sc.Scene3DCanvas           .__init__(self,
                                             overlayList,
                                             displayCtx)



#!/usr/bin/env python
#
# screenshot.py - The ScreenshotAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ScreenshotAction` class, an
:class:`.Action` which can take screenshots of :class:`.CanvasPanel` and
:class:`.PlotPanel` views.


A few stand-alone functions are defined in this module:

.. autosummary::
   :nosignatures:

   screenshot
   plotPanelScreenshot
   canvasPanelScreenshot
"""


import            os
import os.path as op
import            logging

import                      wx
import numpy             as np
import matplotlib.pyplot as plt
import matplotlib.image  as mplimg

from   fsl.utils.platform import platform as fslplatform
import fsl.utils.idle                     as idle
import fsleyes_widgets                    as fwidgets
import fsleyes_widgets.utils.status       as status
import fsl.utils.settings                 as fslsettings
import fsleyes.views.canvaspanel          as canvaspanel
import fsleyes.views.plotpanel            as plotpanel
import fsleyes.plotting.plotcanvas        as plotcanvas

import fsleyes.strings as strings
from . import             base


log = logging.getLogger(__name__)


class ScreenshotAction(base.Action):
    """The ``ScreenshotAction`` is able to save a screenshot of the contents
    of :class:`.CanvasPanel` and :class:`.PlotPanel` views.
    """

    def __init__(self, overlayList, displayCtx, panel):
        """Create a ``ScreenshotAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg panel:       The :class:`.CanvasPanel` or :class:`.PlotPanel` to
                          take a screenshot of. :class:`.PlotCanvas` instances
                          are also accepted.
        """

        base.Action.__init__(
            self, overlayList, displayCtx, self.__doScreenshot)
        self.__panel = panel


    def __doScreenshot(self):
        """Capture a screenshot. Prompts the user to select a file to save
        the screenshot to, and then calls the :func:`screenshot` function.
        """

        lastDirSetting = 'fsleyes.actions.screenshot.lastDir'

        # Ask the user where they want
        # the screenshot to be saved
        fromDir = fslsettings.read(lastDirSetting, os.getcwd())

        # We can get a list of supported output
        # types via a matplotlib figure object
        fig  = plt.figure()
        fmts = fig.canvas.get_supported_filetypes()

        # Default to png if
        # it is available
        if 'png' in fmts:
            fmts = [('png', fmts['png'])] + \
                   [(k, v) for k, v in fmts.items() if k != 'png']
        else:
            fmts = list(fmts.items())

        wildcard = ['[*.{}] {}|*.{}'.format(fmt, desc, fmt)
                    for fmt, desc in fmts]
        wildcard = '|'.join(wildcard)
        filename = 'screenshot'

        dlg = wx.FileDialog(wx.GetTopLevelWindows()[0],
                            message=strings.messages[self, 'screenshot'],
                            wildcard=wildcard,
                            defaultDir=fromDir,
                            defaultFile=filename,
                            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)

        if dlg.ShowModal() != wx.ID_OK:
            return

        filename = dlg.GetPath()

        # Make the dialog go away before
        # the screenshot gets taken
        dlg.Close()
        dlg.Destroy()
        wx.GetApp().Yield()

        # Show an error if the screenshot
        # function raises an error
        doScreenshot = status.reportErrorDecorator(
            strings.titles[  self, 'error'],
            strings.messages[self, 'error'])(screenshot)

        # We do the screenshot asynchronously,
        # to make sure it is performed on
        # the main thread, during idle time
        idle.idle(doScreenshot, self.__panel, filename)

        status.update(strings.messages[self, 'pleaseWait'].format(filename))

        fslsettings.write(lastDirSetting, op.dirname(filename))


def screenshot(panel, filename):
    """Capture a screenshot of the contents of the given :class:`.CanvasPanel`,
    :class:`.PlotPanel`, or :class:`.PlotCanvas`, saving it to the given
    ``filename``.
    """

    if isinstance(panel, canvaspanel.CanvasPanel):
        canvasPanelScreenshot(panel, filename)
    elif isinstance(panel, (plotpanel.PlotPanel, plotcanvas.PlotCanvas)):
        plotPanelScreenshot(panel, filename)


def plotPanelScreenshot(panel, filename):
    """Capture a screenshot of the contents of the given :class:`.PlotPanel`
    or :class:`.PlotPanel`, saving it to the given ``filename``.
    """
    if isinstance(panel, plotpanel.PlotPanel):
        figure = panel.canvas.figure
    elif isinstance(panel, plotcanvas.PlotCanvas):
        figure = panel.figure
    else:
        raise ValueError('Unsupported panel type: {}'.format(type(panel)))

    figure.savefig(filename)


def canvasPanelScreenshot(panel, filename):
    """Capture a screenshot of the contents of the given :class:`.CanvasPanel`
    or :class:`.PlotPanel`, saving it to the given ``filename``.
    """

    # The canvas panel container is the
    # direct parent of the colour bar
    # canvas, and an ancestor of the
    # other GL canvases. So that's the
    # one that we want to take a screenshot
    # of.
    cpanel = panel
    opts   = panel.sceneOpts
    panel  = panel.containerPanel

    # Make a H*W*4 bitmap array (h*w because
    # that's how matplotlib want it). We
    # initialise the bitmap to the current
    # background colour, due to some sizing
    # issues that are discussed in the
    # osxPatch function below.
    bgColour      = np.array(opts.bgColour) * 255
    width, height = panel.GetClientSize().Get()
    data          = np.zeros((height, width, 4), dtype=np.uint8)
    data[:, :, :] = bgColour

    log.debug('Creating bitmap {} * {} for {} screenshot'.format(
        width, height, type(cpanel).__name__))

    # The typical way to get a screen grab of a
    # wx Window is to use a wx.WindowDC, and a
    # wx.MemoryDC, and to 'blit' a region from
    # the window DC into the memory DC. Then we
    # extract the bitmap data and copy it into
    # our array.
    windowDC = wx.WindowDC(panel)
    memoryDC = wx.MemoryDC()

    if fwidgets.wxFlavour() == fwidgets.WX_PHOENIX:
        bmp = wx.Bitmap(width, height)
    else:
        bmp = wx.EmptyBitmap(width, height)

    # Copy the contents of the canvas
    # container to the bitmap
    memoryDC.SelectObject(bmp)
    memoryDC.Blit(
        0,
        0,
        width,
        height,
        windowDC,
        0,
        0)
    memoryDC.SelectObject(wx.NullBitmap)

    rgb = bmp.ConvertToImage().GetData()
    rgb = np.frombuffer(rgb, dtype=np.uint8)

    data[:, :, :3] = rgb.reshape(height, width, 3)

    # OSX and SSh/X11 both have complications
    inSSH = fwidgets.inSSHSession() and not fwidgets.inVNCSession()
    if inSSH or fslplatform.os == 'Darwin':
        data = _patchInCanvases(cpanel, panel, data, bgColour)

    data[:, :,  3] = 255

    try:              fmt = op.splitext(filename)[1][1:]
    except Exception: fmt = None

    mplimg.imsave(filename, data, format=fmt)


def _patchInCanvases(canvasPanel, containerPanel, data, bgColour):
    """Used by the :func:`canvasPanelScreenshot` function.

    For some unknown reason, under OSX and when running over X11/SSH, the
    contents of ``wx.glcanvas.GLCanvas`` instances are not captured by the
    ``WindowDC``/``MemoryDC`` blitting process performed by the
    ``canvasPanelScreenshot`` function - they come out all black.

    So this function manually patches in bitmaps (read from the GL front
    buffer) of each ``GLCanvas`` that is displayed in the canvas panel.
    """

    cpanel = canvasPanel
    panel  = containerPanel

    # Get all the wx GLCanvas instances
    # which are displayed in the panel,
    # including the colour bar canvas
    glCanvases = cpanel.getGLCanvases()
    glCanvases.append(cpanel.colourBarCanvas)

    # Note: These values are not scaled by the
    # display DPI (e.g. if using a retina display).
    # Hackiness below handles this situation.
    totalWidth, totalHeight = panel.GetClientSize().Get()
    absPosx,    absPosy     = panel.GetScreenPosition()

    # Patch in bitmaps for every GL canvas
    for glCanvas in glCanvases:

        # If the colour bar is not displayed,
        # the colour bar canvas will be None
        if glCanvas is None:
            continue

        # Hidden wx objects will
        # still return a size
        if not glCanvas.IsShown():
            continue

        scale         = glCanvas.GetScale()
        width, height = glCanvas.GetScaledSize()
        posx, posy    = glCanvas.GetScreenPosition()

        # make sure canvas position is scaled
        # by the display scaling factor.
        posx -= absPosx
        posy -= absPosy
        posx  = int(posx * scale)
        posy  = int(posy * scale)

        log.debug('Canvas {} position: ({}, {}); size: ({}, {})'.format(
            type(glCanvas).__name__, posx, posy, width, height))

        xstart = posx
        ystart = posy
        xend   = xstart + width
        yend   = ystart + height

        bmp = glCanvas.getBitmap()

        # Under OSX, there seems to be a size/position
        # miscalculation  somewhere, such that if the last
        # canvas is on the hard edge of the parent, the
        # canvas size spills over the parent size by a
        # couple of pixels. If this occurs, I re-size the
        # final bitmap accordingly.
        #
        # n.b. This is why I initialise the bitmap array
        #      to the canvas panel background colour.
        #
        # n.b. This code also (as a byproduct) handles
        #      high-DPI scaling, where the panel size
        #      is in unscaled pixels, but the canvas
        #      sizes/positions are scaled.
        if xend > totalWidth:

            oldWidth    = totalWidth
            totalWidth  = xend
            newData     = np.zeros((totalHeight, totalWidth, 4),
                                   dtype=np.uint8)

            newData[:, :, :]         = bgColour
            newData[:, :oldWidth, :] = data
            data                     = newData

            log.debug('Adjusted bitmap width: {} -> {}'.format(
                oldWidth, totalWidth))

        if yend > totalHeight:

            oldHeight   = totalHeight
            totalHeight = yend
            newData     = np.zeros((totalHeight, totalWidth, 4),
                                   dtype=np.uint8)

            newData[:, :, :]          = bgColour
            newData[:oldHeight, :, :] = data
            data                      = newData

            log.debug('Adjusted bitmap height: {} -> {}'.format(
                oldHeight, totalHeight))

        log.debug('Patching {} in at [{} - {}], [{} - {}]'.format(
            type(glCanvas).__name__, xstart, xend, ystart, yend))

        data[ystart:yend, xstart:xend] = bmp

    return data



#!/usr/bin/env python
#
# group.py - Overlay groups
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OverlayGroup` class, which allows
the display properties of one or more overlays to be linked.
"""


import sys
import copy
import logging


import fsleyes_props                  as props
import fsleyes_widgets.utils.typedict as td


log = logging.getLogger(__name__)


class OverlayGroup(props.HasProperties):
    """An ``OverlayGroup`` is a group of overlays for which the corresponding
    :class:`.Display` and :class:`.DisplayOpts` properties are synchronised.


    The point of the ``OverlayGroup`` is to allow the user to define groups of
    overlays, so he/she can change display properties on the entire group,
    instead of having to change display properties on each overlay one by one.


    Overlays can be added to an ``OverlayGroup`` with the :meth:`addOverlay`,
    and removed with the :meth:`removeOverlay`.


    When an ``OverlayGroup`` is created, it dynamically adds all of the
    properties which could possibly be linked between overlays to itself,
    using the :meth:`props.HasProperties.addProperty` method. When the first
    overlay is added to the group, these group properties are set to the
    display properties of this overlay. Then, the display properties of
    overlays which are subsequently added to the group will be set to the
    group display properties.


    .. note:: Currently, only a subset of display properties are linked
              between the overlays in a group. The properties which are linked
              are hard-coded in the :attr:`_groupBindings` dictionary.

              A possible future *FSLeyes* enhancement will be to allow the
              user to specify which display properties within an
              ``OverlayGroup`` should be linked.
    """


    overlays = props.List()
    """The list of overlays in this ``OverlayGroup``.

    .. warning:: Do not add/remove overlays directly to this list - use the
                :meth:`addOverlay` and :meth:`removeOverlay` methods instead.
    """


    _groupBindings = td.TypeDict({
        'Display'        : [],
        'NiftiOpts'      : ['volume',
                            'volumeDim'],
        'VolumeOpts'     : ['interpolation'],
        'Volume3DOpts'   : ['numSteps',
                            'numInnerSteps',
                            'resolution',
                            'numClipPlanes',
                            'clipMode',
                            'clipPosition',
                            'clipAzimuth',
                            'clipInclination'],
        'LabelOpts'      : [],
        'MeshOpts'       : ['refImage',
                            'coordSpace'],
        'VectorOpts'     : ['suppressX',
                            'suppressY',
                            'suppressZ',
                            'suppressMode'],
        'LineVectorOpts' : [],
        'RGBVectorOpts'  : ['interpolation'],
        'TensorOpts'     : ['lighting',
                            'tensorResolution'],
    })
    """This dictionary defines the properties which are bound across
    :class:`.Display` instances :class:`.DisplayOpts` sub-class instances, for
    overlays which are in the same group.
    """


    def __init__(self, displayCtx, overlayList):
        """Create an ``OverlayGroup``.

        :arg displayCtx:  The :class:`.DisplayContext`.

        :arg overlayList: The :class:`.OverlayList`.
        """

        # Import all of the Display/DisplayOpts
        # classes into the local namespace.
        from fsleyes.displaycontext import (  # noqa
            Display,
            NiftiOpts,
            VolumeOpts,
            Volume3DOpts,
            MaskOpts,
            VectorOpts,
            RGBVectorOpts,
            LineVectorOpts,
            MeshOpts,
            LabelOpts,
            TensorOpts)

        self.__displayCtx  = displayCtx
        self.__overlayList = overlayList
        self.__name        = '{}_{}'.format(type(self).__name__, id(self))

        # This dict is used by the __bindDisplayOpts
        # method to keep track of which group properties
        # have already been given a value
        self.__hasBeenSet  = {}

        # Save a ref to each type so we
        # can look up classes by name.
        self.__typeMappings = {}
        for clsName in OverlayGroup._groupBindings.keys():
            self.__typeMappings[clsName] = locals()[clsName]

        overlayList.addListener('overlays',
                                self.__name,
                                self.__overlayListChanged)

        # Add all of the properties listed
        # in the _groupBindings dict as
        # properties of this OverlayGroup
        # instance.
        for clsName, propNames in OverlayGroup._groupBindings.items():

            cls = self.__typeMappings[clsName]

            for propName in propNames:
                prop = copy.copy(getattr(cls, propName))
                self.addProperty('{}_{}'.format(clsName, propName), prop)
                self.__hasBeenSet[clsName, propName] = False

        # Special case - make sure that the NiftiOpts
        # volume property is not constrained
        self.setAttribute('NiftiOpts_volume', 'maxval', sys.maxsize)


    def __copy__(self):
        """Create a copy of this ``OverlayGroup``.

        A custom copy operator is needed due to the way that
        the :class:`.props.HasProperties` class works.
        """
        return OverlayGroup(self, self.__displayCtx, self.__overlayList)


    def __str__(self):
        """Returns a string representation of this ``OverlayGroup``."""
        return str([str(o) for o in self.overlays])


    def __repr__(self):
        """Returns a string representation of this ``OverlayGroup``."""
        return '[{}]'.format(', '.join([str(o) for o in self.overlays]))


    def __contains__(self, ovl):
        """Returns ``True`` if ``ovl`` is in this ``OverlayGroup``,
        :attr:`False` otherwise.
        """
        return ovl in self.overlays


    def __len__(self):
        """Returns the number of overlays in this ``OverlayGroup``. """
        return len(self.overlays)


    def destroy(self):
        """Must be called when this ``OverlayGroup`` is no longer needed.
        Removes all overlays from the group, and removes property listeners.
        """
        for overlay in list(self.overlays):
            self.removeOverlay(overlay)

        self.__overlayList.removeListener('overlays', self.__name)

        self.__overlayList = None
        self.__displayCtx  = None


    def addOverlay(self, overlay):
        """Add an overlay to this ``OverlayGroup``.

        If this is the first overlay to be added, the properties of this
        ``OverlayGroup`` are set to the overlay display properties. Otherwise,
        the overlay display properties are set to those of this
        ``OverlayGroup``.
        """

        if overlay in self.overlays:
            return

        self.overlays.append(overlay)

        display = self.__displayCtx.getDisplay(overlay)
        opts    = display.opts

        log.debug('Adding overlay {} to group {}'.format(
            overlay.name, self.__name))

        self.__bindDisplayOpts(display)
        self.__bindDisplayOpts(opts)

        display.addListener('overlayType',
                            self.__name,
                            self.__overlayTypeChanged)


    def removeOverlay(self, overlay):
        """Remove the given overlay from this ``OverlayGroup``. """

        if overlay not in self.overlays:
            return

        from . import InvalidOverlayError

        self.overlays.remove(overlay)

        try:
            display = self.__displayCtx.getDisplay(overlay)
        except InvalidOverlayError:
            return

        opts = display.opts

        log.debug('Removing overlay {} from group {}'.format(
            overlay.name, self.__name))

        self.__bindDisplayOpts(display, unbind=True)
        self.__bindDisplayOpts(opts,    unbind=True)

        display.removeListener('overlayType', self.__name)

        # If this was the last overlay of its type,
        # make sure to clear the hasBeenSet flag
        # for all relevent properties; otherwise
        # the properties of the next overlay of
        # this type to be added will be clobbered
        # by the old group values.
        allopts   = [self.__displayCtx.getOpts(o) for o in self.overlays]
        bindProps = OverlayGroup._groupBindings.get(opts,
                                                    allhits=True,
                                                    bykey=True)
        for clsName, propNames in bindProps.items():

            cls     = self.__typeMappings[clsName]
            noftype = len([o for o in allopts if issubclass(type(o), cls)])

            if noftype == 0:
                for propName in propNames:
                    self.__hasBeenSet[clsName, propName] = False


    def __bindDisplayOpts(self, target, unbind=False):
        """Binds or unbinds the properties of the given ``target`` to the
        properties of this ``OverlayGroup``.

        :arg target: A :class:`.Display` or :class:`.DisplayOpts` instance.

        :arg unbind: Set to ``True`` to bind the properties, ``False`` to
                     unbind them.
        """

        # This is the first overlay to be added - the
        # group should inherit its property values
        if len(self.overlays) == 1:
            master, slave = target, self

        # Other overlays are already in the group - the
        # new overlay should inherit the group properties
        else:
            master, slave = self, target

        bindProps = OverlayGroup._groupBindings.get(target,
                                                    allhits=True,
                                                    bykey=True)

        for clsName, propNames in bindProps.items():
            for propName in propNames:

                groupName = '{}_{}'.format(clsName, propName)

                # If the group property has not yet
                # taken on a value, initialise it
                # to the property value being bound.
                #
                # We do this to avoid clobbering
                # property values with un-initialised
                # group property values.
                if not self.__hasBeenSet[clsName, propName]:
                    setattr(self, groupName, getattr(target, propName))
                    self.__hasBeenSet[clsName, propName] = True

                if slave is self:
                    otherName = propName
                    propName  = groupName
                else:
                    otherName = groupName

                slave.bindProps(propName,
                                master,
                                otherName,
                                bindatt=False,
                                unbind=unbind)


    def __overlayListChanged(self, *a):
        """Called when overlays are added/removed to/from the
        :class:`.OverlayList` . Makes sure that the :attr:`overlays` list for
        this group does not contain any overlays that have been removed.
        """
        for ovl in self.overlays:
            if ovl not in self.__overlayList:
                self.removeOverlay(ovl)


    def __overlayTypeChanged(self, value, valid, display, name):
        """This method is called when the :attr:`.Display.overlayType`
        property for an overlay in the group changes.

        It makes sure that the display properties of the new
        :class:`.DisplayOpts` instance are bound to the group properties.
        """
        opts = display.opts
        self.__bindDisplayOpts(opts)



#!/usr/bin/env python
#
# melodicclassificationpanel.py - The MelodicClassificationPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`MelodicClassificationPanel` class, a
*FSLeyes control* panel which allows the user to classify the components
of a :class:`.MelodicImage`.
"""


import os
import os.path as op
import logging

import wx

import fsl.utils.settings            as fslsettings
import fsl.data.volumelabels         as vollabels
import fsl.data.fixlabels            as fixlabels
import fsl.data.image                as fslimage
import fsl.data.melodicimage         as fslmelimage

import fsleyes_props                 as props
import fsleyes_widgets.notebook      as notebook
import fsleyes_widgets.utils.status  as status

import fsleyes.views.canvaspanel     as canvaspanel
import fsleyes.displaycontext        as displaycontext
import fsleyes.colourmaps            as fslcm
import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.autodisplay           as autodisplay
import fsleyes.strings               as strings
from . import componentgrid          as componentgrid
from . import labelgrid              as labelgrid


log = logging.getLogger(__name__)


class MelodicClassificationPanel(ctrlpanel.ControlPanel):
    """The ``MelodicClassificationPanel`` allows the user to view and modify
    classification labels associated with the volumes of an :class:`.Image`,
    most typically the components of a :class:`.MelodicImage` (but any 4D
    image will work).

    A ``MelodicClassificationPanel`` displays two lists:

      - The :class:`.ComponentGrid` contains list of components, and the
        labels associated with each.

      - The :class:`.LabelGrid` contains list of labels, and the
        components associated with each.

    And a handful of buttons which allow the user to:

     - Load a label file

     - Save the current labels to a file

     - Clear/reset the current labels

    Internally, a :class:`.VolumeLabels` object is used to keep track of the
    component - label mappings. The ``VolumeLabels`` instance associated with
    each overlay is stored in the :class:`.OverlayList` via its
    :meth:`.OverlayList.setData` method.
    """


    @staticmethod
    def supportedViews():
        """The ``MelodicClassificationPanel`` is restricted for use with
        :class:`.OrthoPanel`, :class:`.LightBoxPanel` and
        :class:`.Scene3DPanel` viewws.
        """

        return [canvaspanel.CanvasPanel]


    @staticmethod
    def defaultLayout():
        """Returns a dictionary of arguments to be passed to the
        :meth:`.ViewPanel.togglePanel` method.
        """
        return {'location' : wx.RIGHT}


    def __init__(self, parent, overlayList, displayCtx, canvasPanel):
        """Create a ``MelodicClassificationPanel``.

        :arg parent:      The :mod:`wx` parent object.

        :arg overlayList: The :class:`.OverlayList`.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg canvasPanel: The :class:`.CanvasPanel` that owns this
                          classification panel.
        """
        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, canvasPanel)

        self.__disabledText = wx.StaticText(
            self,
            style=(wx.ALIGN_CENTRE_HORIZONTAL |
                   wx.ALIGN_CENTRE_VERTICAL))

        self.__overlay       = None
        self.__canvasPanel   = canvasPanel
        self.__lut           = fslcm.getLookupTable('melodic-classes')

        # If this classification panel has been
        # added to a LightBoxPanel, we add a text
        # annotation to said lightbox panel, to
        # display the labels associated with the
        # currently displayed component.
        self.__textAnnotation = None
        from fsleyes.views.lightboxpanel import LightBoxPanel

        if isinstance(canvasPanel, LightBoxPanel):
            annot = canvasPanel.getCanvas().getAnnotations()
            self.__textAnnotation = annot.text(
                '',
                0.5, 1.0,
                fontSize=30,
                halign='centre',
                valign='top',
                hold=True)

        self.__notebook      = notebook.Notebook(self)
        self.__componentGrid = componentgrid.ComponentGrid(
            self.__notebook,
            overlayList,
            displayCtx,
            canvasPanel.frame,
            self.__lut)

        self.__labelGrid     = labelgrid.LabelGrid(
            self.__notebook,
            overlayList,
            displayCtx,
            canvasPanel.frame,
            self.__lut)

        self.__loadButton  = wx.Button(self)
        self.__saveButton  = wx.Button(self)
        self.__clearButton = wx.Button(self)

        self.__notebook.AddPage(self.__componentGrid,
                                strings.labels[self, 'componentTab'])
        self.__notebook.AddPage(self.__labelGrid,
                                strings.labels[self, 'labelTab'])

        self.__loadButton .SetLabel(strings.labels[self, 'loadButton'])
        self.__saveButton .SetLabel(strings.labels[self, 'saveButton'])
        self.__clearButton.SetLabel(strings.labels[self, 'clearButton'])

        # Things which you don't want shown when
        # a melodic image is not selected should
        # be added to __mainSizer. Things which
        # you always want displayed should be
        # added to __sizer (but need to be laid
        # out w.r.t. __disabledText/__mainSizer)

        self.__mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.__btnSizer  = wx.BoxSizer(wx.HORIZONTAL)
        self.__sizer     = wx.BoxSizer(wx.VERTICAL)

        self.__btnSizer .Add(self.__loadButton,   flag=wx.EXPAND, proportion=1)
        self.__btnSizer .Add(self.__saveButton,   flag=wx.EXPAND, proportion=1)
        self.__btnSizer .Add(self.__clearButton,  flag=wx.EXPAND, proportion=1)

        self.__mainSizer.Add(self.__notebook,     flag=wx.EXPAND, proportion=1)

        self.__sizer    .Add(self.__disabledText, flag=wx.EXPAND, proportion=1)
        self.__sizer    .Add(self.__mainSizer,    flag=wx.EXPAND, proportion=1)
        self.__sizer    .Add(self.__btnSizer,     flag=wx.EXPAND)

        self.SetSizer(self.__sizer)

        self.__loadButton .Bind(wx.EVT_BUTTON, self.__onLoadButton)
        self.__saveButton .Bind(wx.EVT_BUTTON, self.__onSaveButton)
        self.__clearButton.Bind(wx.EVT_BUTTON, self.__onClearButton)

        overlayList.addListener('overlays',
                                self.name,
                                self.__selectedOverlayChanged)
        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__selectedOverlayChanged)

        self.__selectedOverlayChanged()

        self.SetMinSize((400, 100))


    def destroy(self):
        """Must be called when this ``MelodicClassificaiionPanel`` is no longer
        used. Removes listeners, and destroys widgets.
        """

        if self.__textAnnotation is not None:
            annot = self.__canvasPanel.getCanvas().getAnnotations()
            annot.dequeue(self.__textAnnotation, hold=True)

        self.displayCtx .removeListener('selectedOverlay', self.name)
        self.overlayList.removeListener('overlays',        self.name)
        self.__componentGrid.destroy()
        self.__labelGrid    .destroy()

        self.__deregisterOverlay()
        self.__canvasPanel    = None
        self.__textAnnotation = None
        self.__overlay        = None
        self.__lut            = None

        ctrlpanel.ControlPanel.destroy(self)


    def __enable(self, enable=True, message=''):
        """Called internally. Enables/disables this
        ``MelodicClassificationPanel``.
        """

        self.__disabledText.SetLabel(message)

        self.__sizer.Show(self.__disabledText, not enable)
        self.__sizer.Show(self.__mainSizer,    enable)

        self.__saveButton .Enable(enable)
        self.__clearButton.Enable(enable)

        if self.__textAnnotation is not None:
            self.__textAnnotation.enabled = enable

        self.Layout()


    def __deregisterOverlay(self):
        """Called by :meth:`__selectedOverlayChanged`. Deregisters from
        the currently selected overlay.
        """

        from fsleyes.views.lightboxpanel import LightBoxPanel

        overlay        = self.__overlay
        self.__overlay = None

        if overlay is None or \
           not isinstance(self.__canvasPanel, LightBoxPanel):
            return

        volLabels = self.overlayList.getData(overlay, 'VolumeLabels')
        volLabels.deregister(self.name, topic='added')
        volLabels.deregister(self.name, topic='removed')

        try:
            opts = self.displayCtx.getOpts(overlay)
            opts.removeListener('volume', self.name)

        except displaycontext.InvalidOverlayError:
            pass


    def __registerOverlay(self, overlay):
        """Called by :meth:`__selectedOverlayChanged`. Registers with
        the given overlay.

        Returns the :class:`.VolumeLabels` instance associated with the
        overlay (creating it if necessary).
        """

        from fsleyes.views.lightboxpanel import LightBoxPanel

        self.__overlay = overlay

        opts      = self.displayCtx.getOpts(overlay)
        volLabels = self.overlayList.getData(overlay, 'VolumeLabels', None)

        if volLabels is None:
            volLabels = vollabels.VolumeLabels(overlay.shape[3])
            self.overlayList.setData(overlay, 'VolumeLabels', volLabels)

            # Initialse component with an 'Unknown' label
            for i in range(overlay.shape[3]):
                volLabels.addLabel(i, 'Unknown')

        # We only need to listen for volume/label
        # changes if we are in a LightBoxPanel
        if not isinstance(self.__canvasPanel, LightBoxPanel):
            return volLabels

        opts.addListener('volume', self.name, self.__volumeChanged)

        # Whenever the classification labels change,
        # update the text annotation on the canvas.
        # We do this on the idle loop because otherwise,
        # when a new label is added, the LookupTable
        # instance may not have been updated to contain
        # the new label - see ComponentGrid.__onTagAdded.
        for topic in ['added', 'removed']:
            volLabels.register(self.name,
                               self.__labelsChanged,
                               topic=topic,
                               runOnIdle=True)

        self.__updateTextAnnotation()

        return volLabels


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` or
        :attr:`.OverlayList.overlays` changes.

        The overlay is passed to the :meth:`.ComponentGrid.setOverlay`
        and :meth:`.LabelGrid.setOverlay` methods.

        If the newly selected overlay is not a :class:`.MelodicImage`,
        this panel is disabled (via :meth:`__enable`).
        """

        overlay = self.displayCtx.getSelectedOverlay()

        if self.__overlay is overlay:
            return

        self.__deregisterOverlay()

        if (overlay is None)                       or \
           not isinstance(overlay, fslimage.Image) or \
           len(overlay.shape) != 4:
            self.__enable(False, strings.messages[self, 'disabled'])
            volLabels = None

        else:
            volLabels = self.__registerOverlay(overlay)
            self.__enable(True)

        self.__componentGrid.setOverlay(overlay, volLabels)
        self.__labelGrid    .setOverlay(overlay, volLabels)


    def __volumeChanged(self, *a):
        """Called when the :attr:`.NiftiOpts.volume` property of the
        currently selected overlay changes. Calls
        :meth:`__updateTextAnnotation`

        .. note:: This method is only called if the view panel that
                  owns this ``MelodicClassificationPanel`` is a
                  :class:`.LightBoxPanel`.
        """
        self.__updateTextAnnotation()


    def __labelsChanged(self, *a):
        """Called when the :class:`.VolumeLabels` object associated
        with the currently selected overlay changes. Calls
        :meth:`__updateTextAnnotation`

        .. note:: This method is only called if the view panel that
                  owns this ``MelodicClassificationPanel`` is a
                  :class:`.LightBoxPanel`.
        """
        self.__updateTextAnnotation()

        # Label change will not necessarily trigger a
        # canvas refresh, so we manually trigger one
        self.__canvasPanel.Refresh()


    def __updateTextAnnotation(self):
        """Updates a text annotation on the :class:`.LightBoxPanel` canvas to
        display the labels associated with the volume (i.e. the current
        component).
        """

        overlay   = self.__overlay
        opts      = self.displayCtx.getOpts(overlay)
        volLabels = self.overlayList.getData(overlay, 'VolumeLabels')
        labels    = volLabels.getLabels(opts.volume)

        if len(labels) == 0:
            return

        # TODO Currently we're colouring all
        #      labels according to the first
        #      one. You should colour each
        #      label independently, but to do
        #      so, you would need multiple
        #      text annotations (and be able
        #      position them relative to each
        #      other), or the TextAnnotation
        #      class would need to provide the
        #      ability to colour different
        #      portions of the text
        #      independently.
        labels = [volLabels.getDisplayLabel(l) for l in labels]
        colour = self.__lut.getByName(labels[0]).colour

        self.__textAnnotation.text   = ', '.join(labels)
        self.__textAnnotation.colour = colour


    def __onLoadButton(self, ev):
        """Called when the *Load labels* button is pushed.  Prompts the user
        to select a label file to load, then does the following:

        1. If the selected label file refers to the currently selected
           melodic_IC overlay, the labels are applied to the overlay.

        2. If the selected label file refers to a different melodic_IC
           overlay, the user is asked whether they want to load the
           different melodic_IC file (the default), or whether they
           want the labels applied to the existing overlay.

        3. If the selected label file does not refer to any overlay
           (it only contains the bad component list), the user is asked
           whether they want the labels applied to the current melodic_IC
           overlay.

        If the number of labels in the file is less than the number of
        melodic_IC components, the remaining components are labelled
        as unknown. If the number of labels in the file is greater than
        the number of melodic_IC components, an error is shown, and
        nothing is done.
        """

        # The aim of the code beneath the
        # applyLabels function is to load
        # a set of component labels, and
        # to figure out which overlay
        # they should be added to.

        # When it has done this, it calls
        # applyLabels, which applies the
        # loaded labels to the overlay.
        def applyLabels(labelFile, overlay, allLabels, newOverlay):
            # labelFile:  Path to the loaded label file
            # overlay:    Overlay to apply them to
            # allLabels:  Loaded labels (list of (component, [label]) tuples)
            # newOverlay: True if the selected overlay has changed, False
            #             otherwise

            lut       = self.__lut
            volLabels = self.overlayList.getData(overlay, 'VolumeLabels')

            ncomps  = volLabels.numComponents()
            nlabels = len(allLabels)

            # Error: number of labels in the
            # file is greater than the number
            # of components in the overlay.
            if ncomps < nlabels:
                msg   = strings.messages[self, 'wrongNComps'].format(
                    labelFile, overlay.dataSource)
                title = strings.titles[  self, 'loadError']
                wx.MessageBox(msg, title, wx.ICON_ERROR | wx.OK)
                return

            # Number of labels in the file is
            # less than number of components
            # in the overlay - we pad the
            # labels with 'Unknown'
            elif ncomps > nlabels:
                for i in range(nlabels, ncomps):
                    allLabels.append(['Unknown'])

            # Disable notification while applying
            # labels so the component/label grids
            # don't confuse themselves.
            with volLabels.skip(self.__componentGrid.name), \
                 volLabels.skip(self.__labelGrid    .name):

                volLabels.clear()

                for comp, lbls in enumerate(allLabels):
                    for lbl in lbls:
                        volLabels.addLabel(comp, lbl)

                # Make sure a colour in the melodic
                # lookup table exists for all labels
                for label in volLabels.getAllLabels():

                    label    = volLabels.getDisplayLabel(label)
                    lutLabel = lut.getByName(label)

                    if lutLabel is None:
                        log.debug('New melodic classification '
                                  'label: {}'.format(label))
                        lut.new(label, colour=fslcm.randomBrightColour())

            # New overlay was loaded
            if newOverlay:

                # Make sure the new image is selected.
                with props.skip(self.displayCtx,
                                'selectedOverlay',
                                self.name):
                    self.displayCtx.selectOverlay(overlay)

                self.__componentGrid.setOverlay(overlay)
                self.__labelGrid    .setOverlay(overlay)

            # Labels were applied to
            # already selected overlay.
            else:
                self.__componentGrid.refreshTags()
                self.__labelGrid    .refreshTags()

        # If the current overlay is a compatible
        # Image, the open file dialog starting
        # point will be its directory.
        overlay = self.__overlay

        if overlay is not None and overlay.dataSource is not None:
            loadDir = op.dirname(overlay.dataSource)

        # Otherwise it will be the most
        # recent overlay load directory.
        else:
            loadDir = fslsettings.read('loadSaveOverlayDir', os.getcwd())

        # Ask the user to select a label file
        dlg = wx.FileDialog(
            self,
            message=strings.titles[self, 'loadDialog'],
            defaultDir=loadDir,
            style=wx.FD_OPEN)

        # User cancelled the dialog
        if dlg.ShowModal() != wx.ID_OK:
            return

        # Load the specified label file
        filename = dlg.GetPath()
        emsg     = strings.messages[self, 'loadError'].format(filename)
        etitle   = strings.titles[  self, 'loadError']
        try:
            with status.reportIfError(msg=emsg, title=etitle):
                melDir, allLabels = fixlabels.loadLabelFile(filename)
        except Exception:
            return

        # Ok we've got the labels, now
        # we need to figure out which
        # overlay to add them to.

        # If the label file does not refer
        # to a Melodic directory, and the
        # current overlay is a compatible
        # image, apply the labels to the
        # image.
        if overlay is not None and melDir is None:
            applyLabels(filename, overlay, allLabels, False)
            return

        # If the label file refers to a
        # Melodic directory, and the
        # current overlay is a compatible
        # image.
        if overlay is not None and melDir is not None:

            if isinstance(overlay, fslmelimage.MelodicImage):
                overlayDir = overlay.getMelodicDir()
            elif overlay.dataSource is not None:
                overlayDir = op.dirname(overlay.dataSource)
            else:
                overlayDir = 'none'

            # And both the current overlay and
            # the label file refer to the same
            # directory, then we apply the
            # labels to the curent overlay.
            if op.abspath(melDir) == op.abspath(overlayDir):

                applyLabels(filename, overlay, allLabels, False)
                return

            # Otherwise, if the overlay and the
            # label file refer to different
            # directories...

            # Ask the user whether they want to load
            # the image specified in the label file,
            # or apply the labels to the currently
            # selected image.
            dlg = wx.MessageDialog(
                self,
                strings.messages[self, 'diffMelDir'].format(
                    melDir, overlayDir),
                style=wx.ICON_QUESTION | wx.YES_NO | wx.CANCEL)
            dlg.SetYesNoLabels(
                strings.messages[self, 'diffMelDir.labels'],
                strings.messages[self, 'diffMelDir.overlay'])

            response = dlg.ShowModal()

            # User cancelled the dialog
            if response == wx.ID_CANCEL:
                return

            # User chose to load the melodic
            # image specified in the label
            # file. We'll carry on with this
            # processing below.
            elif response == wx.ID_YES:
                pass

            # Apply the labels to the current
            # overlay, even though they are
            # from different analyses.
            else:
                applyLabels(filename, overlay, allLabels, False)
                return

        # If we've reached this far, we are
        # going to attempt to identify the
        # image associated with the label
        # file, load that image, and then
        # apply the labels.

        # The label file does not
        # specify a melodic directory
        if melDir is None:

            msg   = strings.messages[self, 'noMelDir'].format(filename)
            title = strings.titles[  self, 'loadError']
            wx.MessageBox(msg, title, wx.ICON_ERROR | wx.OK)
            return

        # Try loading the melodic_IC image
        # specified in the label file.
        try:
            overlay = fslmelimage.MelodicImage(melDir)

            log.debug('Adding {} to overlay list'.format(overlay))

            with props.skip(self.overlayList, 'overlays',        self.name),\
                 props.skip(self.displayCtx,  'selectedOverlay', self.name):

                self.overlayList.append(overlay)

            if self.displayCtx.autoDisplay:
                autodisplay.autoDisplay(overlay,
                                        self.overlayList,
                                        self.displayCtx)

            fslsettings.write('loadSaveOverlayDir', op.abspath(melDir))

        except Exception as e:

            e     = str(e)
            msg   = strings.messages[self, 'loadError'].format(filename, e)
            title = strings.titles[  self, 'loadError']
            log.debug('Error loading classification file '
                      '({}), ({})'.format(filename, e), exc_info=True)
            wx.MessageBox(msg, title, wx.ICON_ERROR | wx.OK)

        # Apply the loaded labels
        # to the loaded overlay.
        applyLabels(filename, overlay, allLabels, True)


    def __onSaveButton(self, ev):
        """Called when the user pushe the *Save labels* button. Asks the user
        where they'd like the label saved, then saves said labels.
        """

        overlay   = self.displayCtx.getSelectedOverlay()
        volLabels = self.overlayList.getData(overlay, 'VolumeLabels')

        if isinstance(overlay, fslmelimage.MelodicImage):
            defaultDir = overlay.getMelodicDir()
        elif overlay.dataSource is not None:
            defaultDir = op.dirname(overlay.dataSource)
        else:
            defaultDir = None

        dlg = wx.FileDialog(
            self,
            message=strings.titles[self, 'saveDialog'],
            defaultDir=defaultDir,
            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)

        if dlg.ShowModal() != wx.ID_OK:
            return

        filename = dlg.GetPath()
        emsg     = strings.messages[self, 'saveError'].format(filename)
        etitle   = strings.titles[  self, 'saveError']

        with status.reportIfError(msg=emsg, title=etitle, raiseError=False):
            volLabels.save(filename, dirname=defaultDir)


    def __onClearButton(self, ev):
        """Called when the user pushes the *Clear labels* button. Resets
        all of the labels (sets the label for every component to
        ``'Unknown'``).
        """

        overlay   = self.displayCtx.getSelectedOverlay()
        volLabels = self.overlayList.getData(overlay, 'VolumeLabels')

        for c in range(volLabels.numComponents()):

            labels = volLabels.getLabels(c)

            if len(labels) == 1 and labels[0] == 'unknown':
                continue

            volLabels.clearLabels(c)
            volLabels.addLabel(c, 'Unknown')



#!/usr/bin/env python
#
# plotpanel.py - The PlotPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`PlotPanel` and :class:`.OverlayPlotPanel`
classes.  The ``PlotPanel`` class is the base class for all *FSLeyes views*
which display some sort of data plot. The ``OverlayPlotPanel`` is a
``PlotPanel`` which contains some extra logic for displaying plots related to
the currently selected overlay.

The actual plotting logic (using ``matplotilb``) is implemented within the
:class:`.PlotCanvas` class.
"""


import logging

import wx

import fsleyes_widgets.elistbox           as elistbox

import fsleyes.actions                    as actions
import fsleyes.overlay                    as fsloverlay
import fsleyes.colourmaps                 as fslcm
import fsleyes.views.viewpanel            as viewpanel
import fsleyes.plotting                   as plotting
import fsleyes.plotting.plotcanvas        as plotcanvas
import fsleyes.controls.overlaylistpanel  as overlaylistpanel



log = logging.getLogger(__name__)


class PlotPanel(viewpanel.ViewPanel):
    """The ``PlotPanel`` class is the base class for all *FSLeyes views*
    which display some sort of 2D data plot, such as the
    :class:`.TimeSeriesPanel`, and the :class:`.HistogramPanel`.

    .. note:: See also the :class:`OverlayPlotPanel`, which contains extra
              logic for displaying plots related to the currently selected
              overlay, and which is the actual base class used by the
              ``TimeSeriesPanel``, ``HistogramPanel`` and
              ``PowerSpectrumPanel``.


    ``PlotPanel`` uses a :class:`.PlotCanvas`, which in turn uses
    :mod`:matplotlib` for its plotting.  The ``PlotCanvas`` instance used by a
    ``PlotPanel`` can be accessed via the :meth:`canvas` method, which in turn
    can be used to manipulate the plot display settings. The ``matplotlib``
    ``Figure``, ``Axis``, and ``Canvas`` instances can be accessed via the
    ``PlotCanvas`` instance, if they are needed.


    **Sub-class requirements**

    Sub-class implementations of ``PlotPanel`` must do the following:

      1. Call the ``PlotPanel`` constructor.

      2. Define one or more :class:`.DataSeries` sub-classes if needed.

      3. Override the :meth:`draw` method, so it calls the
         :meth:`.PlotCanvas.drawDataSeries` and
         :meth:`.PlotCanvas.drawArtists` methods (:meth:`draw` is passed
         to the ``PlotCanvas`` as a custom ``drawFunc``).

      4. If necessary, override the :meth:`prepareDataSeries` method to
         perform any preprocessing on ``extraSeries`` passed to the
         :meth:`drawDataSeries` method (but not applied to
         :class:`.DataSeries` that have been added to the :attr:`dataSeries`
         list) (:meth:`prepareDataSeries` is passed
         to the ``PlotCanvas`` as a custom ``prepareFunc``).

      5. If necessary, override the :meth:`destroy` method, but make
         sure that the base-class implementation is called.

    **Plot panel actions**

    A number of :mod:`actions` are also provided by the ``PlotPanel`` class:

    .. autosummary::
       :nosignatures:

       screenshot
       importDataSeries
       exportDataSeries
    """


    def controlOptions(self, cpType):
        """Returns some options to be used by :meth:`.ViewPanel.togglePanel`
        for certain control panel types.
        """
        # Tell the overlay list panel to disable
        # all overlays that aren't being plotted.
        #
        # This OverlayPlotPanel will always be
        # notified about a new overlay before
        # this OverlayListPanel, so a DataSeries
        # instance will always have been created
        # by the time the list panel calls this
        # filter function.
        def listFilter(overlay):
            return self.getDataSeries(overlay) is not None

        if cpType is overlaylistpanel.OverlayListPanel:
            return dict(showVis=True,
                        showSave=False,
                        showGroup=False,
                        propagateSelect=True,
                        elistboxStyle=(elistbox.ELB_REVERSE      |
                                       elistbox.ELB_TOOLTIP_DOWN |
                                       elistbox.ELB_NO_ADD       |
                                       elistbox.ELB_NO_REMOVE    |
                                       elistbox.ELB_NO_MOVE),
                        location=wx.LEFT,
                        filterFunc=listFilter)


    def __init__(self,
                 parent,
                 overlayList,
                 displayCtx,
                 frame):
        """Create a ``PlotPanel``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: An :class:`.OverlayList` instance.
        :arg displayCtx:  A :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        """

        viewpanel.ViewPanel.__init__(
            self, parent, overlayList, displayCtx, frame)
        self.__canvas = plotcanvas.PlotCanvas(
            self, self.draw, self.prepareDataSeries)
        self.centrePanel = self.__canvas.canvas


    def destroy(self):
        """Removes some property listeners, and then calls
        :meth:`.ViewPanel.destroy`.
        """
        self.__canvas.destroy()
        self.__canvas = None
        viewpanel.ViewPanel.destroy(self)


    @property
    def canvas(self):
        """Returns a reference to the :class:`.PlotCanvas`. """
        return self.__canvas


    def draw(self, *a):
        """This method must be overridden by ``PlotPanel`` sub-classes.
        Sub-class implementations should call the :meth:`drawDataSeries`
        and meth:`drawArtists` methods.
        """
        raise NotImplementedError()


    def prepareDataSeries(self, ds):
        """Prepares the data from the given :class:`.DataSeries` so it is
        ready to be plotted. Called by the :meth:`__drawOneDataSeries` method
        for any ``extraSeries`` passed to the :meth:`drawDataSeries` method
        (but not applied to :class:`.DataSeries` that have been added to the
        :attr:`dataSeries` list).

        This implementation just returns :class:`.DataSeries.getData` -
        override it to perform any custom preprocessing.
        """
        return ds.getData()


    @actions.action
    def screenshot(self, *a):
        """Prompts the user to select a file name, then saves a screenshot
        of the current plot.

        See the :class:`.ScreenshotAction`.
        """
        from fsleyes.actions.screenshot import ScreenshotAction
        ScreenshotAction(self.overlayList, self.displayCtx, self)()


    @actions.action
    def importDataSeries(self, *a):
        """Imports data series from a text file.

        See the :class:`.ImportDataSeriesAction`.
        """
        from fsleyes.actions.importdataseries import ImportDataSeriesAction
        ImportDataSeriesAction(self.overlayList, self.displayCtx, self)()


    @actions.action
    def exportDataSeries(self, *args, **kwargs):
        """Exports displayed data series to a text file.

        See the :class:`.ExportDataSeriesAction`.
        """
        from fsleyes.actions.exportdataseries import ExportDataSeriesAction
        ExportDataSeriesAction(self.overlayList, self.displayCtx, self)()


class OverlayPlotPanel(PlotPanel):
    """The ``OverlayPlotPanel`` is a :class:`.PlotPanel` which contains
    some extra logic for creating, storing, and drawing :class:`.DataSeries`
    instances for each overlay in the :class:`.OverlayList`.


    **Subclass requirements**

    Sub-classes must:

     1. Implement the :meth:`createDataSeries` method, so it creates a
        :class:`.DataSeries` instance for a specified overlay.

     2. Implement the :meth:`PlotPanel.draw` method so it calls the
        :meth:`.PlotCanvas.drawDataSeries`, passing :class:`.DataSeries`
        instances for all overlays where :attr:`.Display.enabled` is
        ``True``, and the call :meth:`.PlotCanvas.drawArtists` method.

     3. Optionally implement the :meth:`prepareDataSeries` method to
        perform any custom preprocessing.


    **The internal data series store**


    The ``OverlayPlotPanel`` maintains a store of :class:`.DataSeries`
    instances, one for each compatible overlay in the
    :class:`.OverlayList`. The ``OverlayPlotPanel`` manages the property
    listeners that must be registered with each of these ``DataSeries`` to
    refresh the plot.  These instances are created by the
    :meth:`createDataSeries` method, which is implemented by sub-classes. The
    following methods are available to sub-classes, for managing the internal
    store of :class:`.DataSeries` instances:

    .. autosummary::
       :nosignatures:

       getDataSeries
       getDataSeriesToPlot
       clearDataSeries
       updateDataSeries
       addDataSeries
       removeDataSeries


    **Proxy images**


    The ``OverlayPlotPanel`` will replace all :class:`.ProxyImage` instances
    with their base images. This functionality was originally added to support
    the :attr:`.HistogramSeries.showOverlay` functionality - it adds a mask
    image to the :class:`.OverlayList` to display the histogram range.
    Sub-classes may wish to adhere to the same logic (replacing ``ProxyImage``
    instances with their bases)


    **Control panels**


    The :class:`.PlotControlPanel`, :class:`.PlotListPanel`, and
    :class:`.OverlayListPanel` are *FSLeyes control* panels which work with
    the :class:`.OverlayPlotPanel`. The ``PlotControlPanel`` is not intended
    to be used directly - view-specific sub-classes are used instead. These
    panels can be added/removed via the :meth:`.ViewPanel.togglePanel`
    method.


    **Sub-classes**


    The ``OverlayPlotPanel`` is the base class for:

    .. autosummary::
       :nosignatures:

       ~fsleyes.views.timeseriespanel.TimeSeriesPanel
       ~fsleyes.views.histogrampanel.HistogramPanel
       ~fsleyes.views.powerspectrumpanel.PowerSpectrumPanel
    """


    plotColours = {}
    """This dictionary is used to store a collection of ``{overlay : colour}``
    mappings. It is shared across all ``OverlayPlotPanel`` instances, so that
    the same (initial) colour is used for the same overlay, across multiple
    plots.

    See also :attr:`plotStyles`.

    Sub-classes should use the :meth:`getOverlayPlotColour` and
    :meth:`getOverlayPlotStyle` methods to retrieve the initial colour and
    linestyle to use for a given overlay.
    """


    plotStyles = {}
    """This dictionary is used to store a collection of ``{overlay : colour}``
    mappings - it is used in conjunction with :attr:`plotColours`.
    """


    def __init__(self, *args, **kwargs):
        """Create an ``OverlayPlotPanel``.

        :arg initialState: Must be passed as a keyword argument. Allows you to
                           specify the initial enabled/disabled state for each
                           overlay. See :meth:`updateDataSeries`. If not
                           provided, only the data series for the currently
                           selected overlay is shown (if possible).

        All other argumenst are passed through to :meth:`PlotPanel.__init__`.
        """

        initialState = kwargs.pop('initialState', None)

        PlotPanel.__init__(self, *args, **kwargs)

        self.__name = 'OverlayPlotPanel_{}'.format(self.name)

        # The dataSeries attribute is a dictionary of
        #
        #   {overlay : DataSeries}
        #
        # mappings, containing a DataSeries instance for
        # each compatible overlay in the overlay list.
        #
        # refreshProps is a dict of
        #
        #   {overlay : ([targets], [propNames]}
        #
        # mappings, containing the target instances and
        # properties which, when those properties change,
        # need to trigger a refresh of the plot.
        #
        # refreshCounts is a dict of:
        #
        #   {target, propName : dscount}

        # mappings, containing all targets and
        # associated property names on which a listener
        # is currently registered, and the count of
        # DataSeries instances which are interested
        # in them.
        self.__dataSeries    = {}
        self.__refreshProps  = {}
        self.__refreshCounts = {}

        # Pre-generated default colours and line
        # styles to use - see plotColours, plotStyles,
        # getOverlayPlotColour, and getOverlayPlotStyle
        lut    = fslcm.getLookupTable('paul_tol_accessible')
        styles = plotting.DataSeries.lineStyle.getChoices()
        limit  = min(len(lut), len(styles))
        self.__defaultColours = [l.colour for l in lut[   :limit]]
        self.__defaultStyles  = [s        for s in styles[:limit]]

        self.canvas     .addListener('dataSeries',
                                     self.__name,
                                     self.__dataSeriesChanged)
        self.overlayList.addListener('overlays',
                                     self.__name,
                                     self.__overlayListChanged)

        self.__overlayListChanged(initialState=initialState)
        self.__dataSeriesChanged()


    def destroy(self):
        """Must be called when this ``OverlayPlotPanel`` is no longer needed.
        Removes some property listeners, and calls :meth:`PlotPanel.destroy`.
        """
        self.overlayList.removeListener('overlays',   self.__name)
        self.canvas     .removeListener('dataSeries', self.__name)

        for overlay in list(self.__dataSeries.keys()):
            self.clearDataSeries(overlay)

        self.__dataSeries    = None
        self.__refreshProps  = None
        self.__refreshCounts = None

        PlotPanel.destroy(self)


    def getDataSeriesToPlot(self):
        """Convenience method which returns a list of overlays which have
        :class:`.DataSeries` that should be plotted.
        """

        overlays = self.overlayList[:]

        # Display.enabled
        overlays = [o for o in overlays
                    if self.displayCtx.getDisplay(o).enabled]

        # Replace proxy images
        overlays = [o.getBase() if isinstance(o, fsloverlay.ProxyImage)
                    else o for o in overlays]

        # Have data series
        dss = [self.getDataSeries(o) for o in overlays]
        dss = [ds for ds in dss if ds is not None]

        # Gather any extra time series
        # associated with the base time
        # series objects.
        for i, ds in enumerate(list(reversed(dss))):

            extras = ds.extraSeries()
            dss    = dss[:i + 1] + extras + dss[i + 1:]

            # If a base time series is disabled,
            # its additional ones should also
            # be disabled
            for eds in extras:
                eds.enabled = ds.enabled

        # Remove duplicates
        unique = []
        for ds in dss:
            if ds not in unique:
                unique.append(ds)

        return unique


    def getDataSeries(self, overlay):
        """Returns the :class:`.DataSeries` instance associated with the
        specified overlay, or ``None`` if there is no ``DataSeries`` instance.
        """

        if isinstance(overlay, fsloverlay.ProxyImage):
            overlay = overlay.getBase()

        return self.__dataSeries.get(overlay)


    def getOverlayPlotColour(self, overlay):
        """Returns an initial colour to use for plots associated with the
        given overlay. If a colour is present in the  :attr:`plotColours`
        dictionary, it is returned. Otherwise a random colour is generated,
        added to ``plotColours``, and returned.
        """

        if isinstance(overlay, fsloverlay.ProxyImage):
            overlay = overlay.getBase()

        colour = self.plotColours.get(overlay)

        if colour is None:
            idx    = len(self.plotColours) % len(self.__defaultColours)
            colour = self.__defaultColours[idx]
            self.plotColours[overlay] = colour

        return colour


    def getOverlayPlotStyle(self, overlay):
        """Returns an initial line style to use for plots associated with the
        given overlay. If a colour is present in the  :attr:`plotStyles`
        dictionary, it is returned. Otherwise a line style is generated,
        added to ``plotStyles``, and returned.

        The format of the returned line style is suitable for use with the
        ``linestyle`` argument of the ``matplotlib`` ``plot`` functions.
        """

        if isinstance(overlay, fsloverlay.ProxyImage):
            overlay = overlay.getBase()

        style = self.plotStyles.get(overlay)

        if style is None:
            idx   = len(self.plotStyles) % len(self.__defaultStyles)
            style = self.__defaultStyles[idx]
            self.plotStyles[overlay] = style

        return style


    @actions.action
    def addDataSeries(self):
        """Every :class:`.DataSeries` which is currently plotted, and has not
        been added to the :attr:`.PlotCanvas.dataSeries` list, is added to said
        list.
        """

        # Get all the DataSeries objects which
        # have been drawn, and are not in the
        # dataSeries list.
        toAdd = self.canvas.getDrawnDataSeries()
        toAdd = [d[0] for d in toAdd if d[0] not in self.canvas.dataSeries]

        if len(toAdd) == 0:
            return

        # Replace each DataSeries instance with a copy.
        # This is necessary because some DataSeries
        # sub-classes have complicated behaviour (e.g.
        # changing their data when some properties
        # change). But we just want to 'freeze' the
        # data as it is currently shown. So we create
        # a dumb copy.
        for i, ds  in enumerate(toAdd):

            copy           = plotting.DataSeries(ds.overlay,
                                                 self.overlayList,
                                                 self.displayCtx,
                                                 self.canvas)
            toAdd[i]       = copy

            copy.alpha     = ds.alpha
            copy.lineWidth = ds.lineWidth
            copy.lineStyle = ds.lineStyle
            copy.label     = ds.label
            copy.colour    = ds.colour

            # We have to re-generate the data,
            # because the x/y data returned by
            # the getDrawnDataSeries method
            # above may have had post-processing
            # applied to it (e.g. smoothing)
            xdata, ydata = self.prepareDataSeries(ds)

            copy.setData(xdata, ydata)

            # This is disgraceful. It wasn't too bad
            # when this function was defined in the
            # PlotListPanel class, but is a horrendous
            # hack now that it is defined here in the
            # PlotPanel class.
            #
            # At some stage I will remove this offensive
            # code, and figure out a more robust system
            # for appending this metadata to DataSeries
            # instances.
            #
            # When the user selects a data series in
            # the list, we want to change the selected
            # overlay/location/volume/etc to the
            # properties associated with the data series.
            # So here we're adding some attributes to
            # each data series instance so that the
            # PlotListPanel.__onListSelect method can
            # update the display properties.
            opts = self.displayCtx.getOpts(ds.overlay)
            if isinstance(ds, (plotting.MelodicTimeSeries,
                               plotting.MelodicPowerSpectrumSeries)):
                copy._volume = opts.volume

            elif isinstance(ds, plotting.VoxelDataSeries):
                copy._location = opts.getVoxel()

        self.canvas.dataSeries.extend(toAdd)


    @actions.action
    def removeDataSeries(self, *a):
        """Removes the most recently added :class:`.DataSeries` from the
        :attr:`.PlotCanvas.dataSeries`.
        """
        if len(self.canvas.dataSeries) > 0:
            self.canvas.dataSeries.pop()


    def createDataSeries(self, overlay):
        """This method must be implemented by sub-classes. It must create and
        return a :class:`.DataSeries` instance for the specified overlay.


        .. note:: Sub-class implementations should set the
                  :attr:`.DataSeries.colour` property to that returned by
                  the :meth:`getOverlayPlotColour` method, and the
                  :attr:`.DataSeries.lineStyle` property to that returned by
                  the :meth:`getOverlayPlotStyle` method


        Different ``DataSeries`` types need to be re-drawn when different
        properties change. For example, a :class:`.VoxelTimeSeries`` instance
        needs to be redrawn when the :attr:`.DisplayContext.location` property
        changes, whereas a :class:`.MelodicTimeSeries` instance needs to be
        redrawn when the :attr:`.VolumeOpts.volume` property changes.


        Therefore, in addition to creating and returning a ``DataSeries``
        instance for the given overlay, sub-class implementations must also
        specify the properties which affect the state of the ``DataSeries``
        instance. These must be specified as two lists:

         - the *targets* list, a list of objects which own the dependant
           properties (e.g. the :class:`.DisplayContext` or
           :class:`.VolumeOpts` instance).

         - The *properties* list, a list of names, each specifying the
           property on the corresponding target.

        This method must therefore return a tuple containing:

          - A :class:`.DataSeries` instance, or ``None`` if the overlay
            is incompatible.
          - A list of *target* instances.
          - A list of *property names*.

        The target and property name lists must have the same length.
        """
        raise NotImplementedError('createDataSeries must be '
                                  'implemented by sub-classes')


    def clearDataSeries(self, overlay):
        """Destroys the internally cached :class:`.DataSeries` for the given
        overlay.
        """

        if isinstance(overlay, fsloverlay.ProxyImage):
            overlay = overlay.getBase()

        ds                 = self.__dataSeries  .pop(overlay, None)
        targets, propNames = self.__refreshProps.pop(overlay, ([], []))

        if ds is not None:

            log.debug('Destroying {} for {}'.format(
                type(ds).__name__, overlay))

            for propName in ds.redrawProperties():
                ds.removeListener(propName, self.__name)
            ds.destroy()

        for t, p in zip(targets, propNames):
            count = self.__refreshCounts[t, p]

            if count - 1 == 0:
                self.__refreshCounts.pop((t, p))
                t.removeListener(p, self.__name)
            else:
                self.__refreshCounts[t, p] = count - 1


    def updateDataSeries(self, initialState=None):
        """Makes sure that a :class:`.DataSeries` instance has been created
        for every compatible overlay, and that property listeners are
        correctly registered, so the plot can be refreshed when needed.

        :arg initialState: If provided, must be a ``dict`` of ``{ overlay :
                           bool }`` mappings, specifying the initial value
                           of the :attr:`.DataSeries.enabled` property for
                           newly created instances. If not provided, only
                           the data series for the currently selected
                           overlay (if it has been newly added) is initially
                           enabled.
        """

        # Default to showing the
        # currently selected overlay
        if initialState is None:
            if len(self.overlayList) > 0:
                initialState = {self.displayCtx.getSelectedOverlay() : True}
            else:
                initialState = {}

        # Make sure that a DataSeries
        # exists for every compatible overlay
        newOverlays = []
        for ovl in self.overlayList:

            if ovl in self.__dataSeries:
                continue

            if isinstance(ovl, fsloverlay.ProxyImage):
                continue

            ds, refreshTargets, refreshProps = self.createDataSeries(ovl)
            display                          = self.displayCtx.getDisplay(ovl)

            if ds is None:

                # "Disable" overlays which don't have any data
                # to plot. We do this mostly so the overlay
                # appears greyed out in the OverlayListPanel.
                display.enabled = False
                continue

            # Display.enabled == DataSeries.enabled
            ds.bindProps('enabled', display)

            ds.enabled = initialState.get(ovl, False)

            log.debug('Created {} for overlay {} (enabled: {})'.format(
                type(ds).__name__, ovl, ds.enabled))

            newOverlays.append(ovl)

            self.__dataSeries[  ovl] = ds
            self.__refreshProps[ovl] = (refreshTargets, refreshProps)

        # Make sure that property listeners are
        # registered all of these overlays
        for overlay in newOverlays:

            targets, propNames = self.__refreshProps.get(overlay, (None, None))

            if targets is None:
                continue

            ds = self.__dataSeries[overlay]

            for propName in ds.redrawProperties():
                ds.addListener(propName,
                               self.__name,
                               self.canvas.asyncDraw,
                               overwrite=True)

            for target, propName in zip(targets, propNames):

                count = self.__refreshCounts.get((target, propName), 0)
                self.__refreshCounts[target, propName] = count + 1

                if count == 0:

                    log.debug('Adding listener on {}.{} for {} data '
                              'series'.format(type(target).__name__,
                                              propName,
                                              overlay))

                    target.addListener(propName,
                                       self.__name,
                                       self.canvas.asyncDraw,
                                       overwrite=True)


    def __dataSeriesChanged(self, *a):
        """Called when the :attr:`.PlotCanvas.dataSeries` list
        changes. Enables/disables the :meth:`removeDataSeries` action
        accordingly.
        """
        self.removeDataSeries.enabled = len(self.canvas.dataSeries) > 0


    def __overlayListChanged(self, *a, **kwa):
        """Called when the :class:`.OverlayList` changes. Makes sure that
        there are no :class:`.DataSeries` instances in the
        :attr:`.PlotCanvas.dataSeries` list, or in the internal cache, which
        refer to overlays that no longer exist.

        :arg initialState: Must be passed as a keyword argument. If provided,
                           passed through to the :meth:`updateDataSeries`
                           method.
        """

        initialState = kwa.get('initialState', None)

        for ds in list(self.canvas.dataSeries):
            if ds.overlay is not None and ds.overlay not in self.overlayList:
                self.canvas.dataSeries.remove(ds)
                ds.destroy()

        for overlay in list(self.__dataSeries.keys()):
            if overlay not in self.overlayList:
                self.clearDataSeries(overlay)

        for overlay in self.overlayList:
            display = self.displayCtx.getDisplay(overlay)

            # PlotPanels use the Display.enabled property
            # to toggle on/off overlay plots. We don't want
            # this to interfere with CanvasPanels, which
            # use Display.enabled to toggle on/off overlays.
            display.detachFromParent('enabled')

        self.updateDataSeries(initialState=initialState)
        self.canvas.asyncDraw()



#!/usr/bin/env python
#
# program.py - The GLSLShader class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLSLShader` class, which encapsulates
a GLSL shader program comprising a vertex shader and a fragment shader.
"""


import logging

import jinja2                         as j2
import numpy                          as np
import OpenGL.GL                      as gl
import OpenGL.raw.GL._types           as gltypes
import OpenGL.GL.ARB.instanced_arrays as arbia

import fsl.utils.memoize as memoize
from . import               parse


log = logging.getLogger(__name__)


GLSL_ATTRIBUTE_TYPES = {
    'bool'          : (gl.GL_BOOL,  1),
    'int'           : (gl.GL_INT,   1),
    'float'         : (gl.GL_FLOAT, 1),
    'vec2'          : (gl.GL_FLOAT, 2),
    'vec3'          : (gl.GL_FLOAT, 3),
    'vec4'          : (gl.GL_FLOAT, 4)
}
"""This dictionary contains mappings between GLSL data types, and their
corresponding GL types and sizes.
"""


class GLSLShader(object):
    """The ``GLSLShader`` class encapsulates information and logic about
    a GLSL 1.20 shader program, comprising a vertex shader and a fragment
    shader. It provides methods to set shader attribute and uniform values,
    to configure attributes, and to load/unload the program. Furthermore,
    the ``GLSLShader`` makes sure that all uniform and attribute variables
    are converted to the appropriate type. The following methods are available
    on a ``GLSLShader``:


    .. autosummary::
       :nosignatures:

       load
       unload
       destroy
       loadAtts
       unloadAtts
       set
       setAtt
       setIndices


    Typical usage of a ``GLSLShader`` will look something like the
    following::

        vertSrc = 'vertex shader source'
        fragSrc = 'fragment shader source'

        program = GLSLShader(vertSrc, fragSrc)

        # Load the program
        program.load()

        # Set some uniform values
        program.set('lighting', True)
        program.set('lightPos', [0, 0, -1])

        # Create and set vertex attributes
        vertices, normals = createVertices()

        program.setAtt('vertex', vertices)
        program.setAtt('normal', normals)

        # Load the attributes
        program.loadAtts()

        # Draw the scene
        gl.glDrawArrays(gl.GL_TRIANGLES, 0, len(vertices))

        # Clear the GL state
        program.unload()
        program.unloadAtts()


        # Delete the program when
        # we no longer need it
        program.destroy()
    """


    def __init__(self, vertSrc, fragSrc, indexed=False, constants=None):
        """Create a ``GLSLShader``.

        The source is passed through ``jinja2``, replacing any expressions on
        the basis of ``constants``.

        :arg vertSrc:   String containing vertex shader source code.

        :arg fragSrc:   String containing fragment shader source code.

        :arg indexed:   If ``True``, it is assumed that the vertices processed
                        by this shader program will be drawn using an index
                        array.  A vertex buffer object is created to store
                        vertex indices - this buffer is expected to be
                        populated via the :meth:`setIndices` method.

        :arg constants: Key-value pairs to be used when passing the source
                        through ``jinja2``.
        """

        if constants is None:
            constants = {}

        vertSrc      = j2.Template(vertSrc).render(**constants)
        fragSrc      = j2.Template(fragSrc).render(**constants)
        self.program = self.__compile(vertSrc, fragSrc)

        vertDecs  = parse.parseGLSL(vertSrc)
        fragDecs  = parse.parseGLSL(fragSrc)
        vertUnifs = vertDecs['uniform']
        vertAtts  = vertDecs['attribute']
        fragUnifs = fragDecs['uniform']

        if len(vertUnifs)  > 0: vuNames, vuTypes, vuSizes = zip(*vertUnifs)
        else:                   vuNames, vuTypes, vuSizes = [], [], []
        if len(vertAtts)  > 0:  vaNames, vaTypes, vaSizes = zip(*vertAtts)
        else:                   vaNames, vaTypes, vaSizes = [], [], []
        if len(fragUnifs) > 0:  fuNames, fuTypes, fuSizes = zip(*fragUnifs)
        else:                   fuNames, fuTypes, fuSizes = [], [], []

        allTypes = {}
        allSizes = {}

        for n, t in zip(vuNames, vuTypes): allTypes[n] = t
        for n, t in zip(vaNames, vaTypes): allTypes[n] = t
        for n, t in zip(fuNames, fuTypes): allTypes[n] = t
        for n, s in zip(vuNames, vuSizes): allSizes[n] = s
        for n, s in zip(vaNames, vaSizes): allSizes[n] = s
        for n, s in zip(fuNames, fuSizes): allSizes[n] = s

        # Remove duplicate uniform definitions
        # between the vertex/fragment shader -
        # they only need to be set once.
        vertUnifs = [vu for vu in vertUnifs if vu not in fragUnifs]

        self.vertUniforms    = vuNames
        self.vertAttributes  = vaNames
        self.fragUniforms    = fuNames

        self.vertAttDivisors = {}

        self.types     = allTypes
        self.sizes     = allSizes
        self.positions = self.__getPositions(self.program,
                                             self.vertAttributes,
                                             self.vertUniforms,
                                             self.fragUniforms)

        # Buffers for vertex attributes
        self.buffers = {}

        for att in self.vertAttributes:
            self.buffers[att] = gl.glGenBuffers(1)

        if indexed: self.indexBuffer = gl.glGenBuffers(1)
        else:       self.indexBuffer = None

        log.debug('{}.init({})'.format(type(self).__name__, id(self)))


    def __del__(self):
        """Prints a log message. """
        if log:
            log.debug('{}.del({})'.format(type(self).__name__, id(self)))


    def load(self):
        """Loads this ``GLSLShader`` into the GL state.
        """
        gl.glUseProgram(self.program)


    def loadAtts(self):
        """Binds all of the shader program ``attribute`` variables - you
        must set the data for each attribute via :meth:`setAtt` before
        calling this method.
        """
        for att in self.vertAttributes:

            aPos           = self.positions[          att]
            aType          = self.types[              att]
            aBuf           = self.buffers[            att]
            aDivisor       = self.vertAttDivisors.get(att)
            glType, glSize = GLSL_ATTRIBUTE_TYPES[aType]

            gl.glBindBuffer(gl.GL_ARRAY_BUFFER, aBuf)
            gl.glEnableVertexAttribArray(aPos)
            gl.glVertexAttribPointer(    aPos,
                                         glSize,
                                         glType,
                                         gl.GL_FALSE,
                                         0,
                                         None)

            if aDivisor is not None:
                arbia.glVertexAttribDivisorARB(aPos, aDivisor)

        if self.indexBuffer is not None:
            gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self.indexBuffer)


    def unloadAtts(self):
        """Disables all vertex attributes, and unbinds associated vertex buffers.
        """
        for att in self.vertAttributes:
            gl.glDisableVertexAttribArray(self.positions[att])

            pos     = self.positions[          att]
            divisor = self.vertAttDivisors.get(att)

            if divisor is not None:
                arbia.glVertexAttribDivisorARB(pos, 0)

        if self.indexBuffer is not None:
            gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, 0)

        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)


    def unload(self):
        """Unloads the GL shader program. """
        gl.glUseProgram(0)


    def destroy(self):
        """Deletes all GL resources managed by this ``GLSLShader``. """
        gl.glDeleteProgram(self.program)

        for buf in self.buffers.values():
            gl.glDeleteBuffers(1, gltypes.GLuint(buf))
        self.program = None


    @memoize.Instanceify(memoize.skipUnchanged)
    def set(self, name, value, size=None):
        """Sets the value for the specified GLSL ``uniform`` variable.

        The ``GLSLShader`` keeps a copy of the value of every uniform, to
        avoid unnecessary GL calls.


        .. note:: This method is decorated by the
                  :func:`.memoize.skipUnchanged` decorator, which returns
                  ``True`` if the value was changed, ``False`` otherwise.
        """

        vPos  = self.positions[name]
        vType = self.types[    name]
        vSize = self.sizes[    name]

        if size is None:
            size = 1

        setfunc = getattr(self, '_uniform_{}'.format(vType), None)

        if size > vSize:
            raise RuntimeError('Specified size ({}) is greater than '
                               'uniform size {} ({})'.format(
                                   size, name, vSize))

        if setfunc is None:
            raise RuntimeError('Unsupported shader program '
                               'type: {}'.format(vType))

        log.debug('Setting shader variable: {}({})[{}] = {}'.format(
            vType, size, name, value))

        setfunc(vPos, value, size)


    def setAtt(self, name, value, divisor=None):
        """Sets the value for the specified GLSL ``attribute`` variable.

        :arg divisor: If specified, this value is used as a divisor for this
                      attribute via the ``glVetexAttribDivisor`` function.

        .. note:: If a ``divisor`` is specified, the OpenGL
                  ``ARB_instanced_arrays`` extension must be
                  available.
        """

        aType    = self.types[  name]
        aBuf     = self.buffers[name]

        castfunc = getattr(self, '_attribute_{}'.format(aType), None)

        if castfunc is None:
            raise RuntimeError('Unsupported shader program '
                               'type: {}'.format(aType))

        value = castfunc(value)

        log.debug('Setting shader attribute: {}({}): {}'.format(
            aType, name, value.shape))

        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, aBuf)
        gl.glBufferData(gl.GL_ARRAY_BUFFER,
                        value.nbytes,
                        value,
                        gl.GL_STATIC_DRAW)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)

        if divisor is not None:
            self.vertAttDivisors[name] = divisor


    def setIndices(self, indices):
        """If an index array is to be used by this ``GLSLShader`` (see the
        ``indexed`` argument to :meth:`__init__`), the index array may be set
        via this method.
        """

        if self.indexBuffer is None:
            raise RuntimeError('Shader program was not '
                               'configured with index support')

        indices = np.array(indices, dtype=np.uint32)

        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER,
                        self.indexBuffer)
        gl.glBufferData(gl.GL_ELEMENT_ARRAY_BUFFER,
                        indices.nbytes,
                        indices,
                        gl.GL_STATIC_DRAW)
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, 0)


    def __getPositions(self, shaders, vertAtts, vertUniforms, fragUniforms):
        """Gets the position indices for all vertex shader attributes,
        uniforms, and fragment shader uniforms for the given shader
        programs.

        :arg shaders:      Reference to the compiled shader program.
        :arg vertAtts:     List of attributes required by the vertex shader.
        :arg vertUniforms: List of uniforms required by the vertex shader.
        :arg fragUniforms: List of uniforms required by the fragment shader.

        :returns:  A dictionary of ``{name : position}`` mappings.
        """

        import OpenGL.GL as gl

        shaderVars = {}

        for v in vertUniforms:
            shaderVars[v] = gl.glGetUniformLocation(shaders, v)

        for v in vertAtts:
            shaderVars[v] = gl.glGetAttribLocation(shaders, v)

        for v in fragUniforms:
            if v in shaderVars:
                continue
            shaderVars[v] = gl.glGetUniformLocation(shaders, v)

        return shaderVars


    def __compile(self, vertShaderSrc, fragShaderSrc):
        """Compiles and links the OpenGL GLSL vertex and fragment shader
        programs, and returns a reference to the resulting program. Raises
        an error if compilation/linking fails.

        .. note:: I'm explicitly not using the PyOpenGL
                  :func:`OpenGL.GL.shaders.compileProgram` function, because
                  it attempts to validate the program after compilation, which
                  fails due to texture data not being bound at the time of
                  validation.
        """

        # vertex shader
        vertShader = gl.glCreateShader(gl.GL_VERTEX_SHADER)
        gl.glShaderSource(vertShader, vertShaderSrc)
        gl.glCompileShader(vertShader)
        vertResult = gl.glGetShaderiv(vertShader, gl.GL_COMPILE_STATUS)

        if vertResult != gl.GL_TRUE:
            raise RuntimeError('{}'.format(gl.glGetShaderInfoLog(vertShader)))

        # fragment shader
        fragShader = gl.glCreateShader(gl.GL_FRAGMENT_SHADER)
        gl.glShaderSource(fragShader, fragShaderSrc)
        gl.glCompileShader(fragShader)
        fragResult = gl.glGetShaderiv(fragShader, gl.GL_COMPILE_STATUS)

        if fragResult != gl.GL_TRUE:
            raise RuntimeError('{}'.format(gl.glGetShaderInfoLog(fragShader)))

        # link all of the shaders!
        program = gl.glCreateProgram()
        gl.glAttachShader(program, vertShader)
        gl.glAttachShader(program, fragShader)

        gl.glLinkProgram(program)

        gl.glDeleteShader(vertShader)
        gl.glDeleteShader(fragShader)

        linkResult = gl.glGetProgramiv(program, gl.GL_LINK_STATUS)

        if linkResult != gl.GL_TRUE:
            raise RuntimeError('{}'.format(gl.glGetProgramInfoLog(program)))

        return program


    def _attribute_bool(self, val):
        return np.array(val, dtype=np.bool, copy=False)


    def _attribute_int(self, val):
        return np.array(val, dtype=np.int32, copy=False)


    def _attribute_float(self, val):
        return np.array(val, dtype=np.float32, copy=False)


    def _attribute_vec2(self, val):
        return np.array(val, dtype=np.float32, copy=False).ravel('C')


    def _attribute_vec3(self, val):
        return np.array(val, dtype=np.float32, copy=False).ravel('C')


    def _attribute_vec4(self, val):
        return np.array(val, dtype=np.float32, copy=False).ravel('C')


    def _uniform_bool(self, pos, val, size):
        val = np.array(val, dtype=np.int32)
        gl.glUniform1iv(pos, size, val)


    def _uniform_int(self, pos, val, size):
        val = np.array(val, dtype=np.int32)
        gl.glUniform1iv(pos, size, val)


    def _uniform_float(self, pos, val, size):
        val = np.array(val, dtype=np.float32)
        gl.glUniform1fv(pos, size, val)


    def _uniform_vec2(self, pos, val, size):
        val = np.array(val, dtype=np.float32, copy=False).ravel('C')
        gl.glUniform2fv(pos, size, val)


    def _uniform_vec3(self, pos, val, size):
        val = np.array(val, dtype=np.float32, copy=False).ravel('C')
        gl.glUniform3fv(pos, size, val)


    def _uniform_vec4(self, pos, val, size):
        val = np.array(val, dtype=np.float32, copy=False).ravel('C')
        gl.glUniform4fv(pos, size, val)


    def _uniform_mat2(self, pos, val, size):
        val = np.array(val, dtype=np.float32, copy=False).ravel('C')
        gl.glUniformMatrix2fv(pos, size, True, val)


    def _uniform_mat3(self, pos, val, size):
        val = np.array(val, dtype=np.float32, copy=False).ravel('C')
        gl.glUniformMatrix3fv(pos, size, True, val)


    def _uniform_mat4(self, pos, val, size):
        val = np.array(val, dtype=np.float32, copy=False).ravel('C')
        gl.glUniformMatrix4fv(pos, 1, True, val)


    def _uniform_sampler1D(self, pos, val, size):
        val = np.array(val, dtype=np.int32, copy=False).ravel('C')
        gl.glUniform1iv(pos, size, val)


    def _uniform_sampler2D(self, pos, val, size):
        val = np.array(val, dtype=np.int32, copy=False).ravel('C')
        gl.glUniform1iv(pos, size, val)


    def _uniform_sampler3D(self, pos, val, size):
        val = np.array(val, dtype=np.int32, copy=False).ravel('C')
        gl.glUniform1iv(pos, size, val)



#!/usr/bin/env python
#
# gltensor_funcs.py - OpenGL2.1 functions used by the GLTensor class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions which are used by the :class:`.GLTensor`
class for rendering :class:`.DTIFitTensor` overlays, and compatible
:class:`.Image` overlays in an OpenGL 2.1 compatible manner.


The rendering code makes use of the OpenGL ``ARB_draw_instanced`` extension
so that voxel coordinates do not need to be repeated for every vertex of
a single tensor.


For each voxel, the vertices of a unit sphere are passed to the ``gltensor``
vertex shader, which looks up the eigenvectors and values for the voxel, and
transforms the sphere accordingly.


If the :attr:`.VectorOpts.colourImage` property is not set, the ``glvector``
fragment shader is used to colour the tensors. Otherwise, the ``glvolume``
fragment shader is used to colour the tensors according to the specified
``colourImage``. The functions in the :mod:`.gl21.glvector_funcs` module
are used to manage the fragment shader.
"""


import numpy                        as np
import numpy.linalg                 as npla
import OpenGL.GL                    as gl
import OpenGL.GL.ARB.draw_instanced as arbdi

import fsl.transform.affine  as affine
import fsleyes.gl.routines   as glroutines
from . import                   glvector_funcs


def init(self):
    """Calls :func:`compileShaders` and :func:`updateShaderState`.
    """

    self.shader = None

    compileShaders(self)
    updateShaderState(self)


def destroy(self):
    """Deletes the :class:`.GLSLShader`. """

    if self.shader is not None:
        self.shader.destroy()
    self.shader = None


def compileShaders(self):
    """Creates a :class:`.GLSLShader` for drawing this ``GLTensor``. This is
    done via a call to :func:`.gl21.glvector_funcs.compileShaders`.
    """
    self.shader = glvector_funcs.compileShaders(self, 'gltensor', indexed=True)


def updateShaderState(self):
    """Updates the state of the vertex and fragment shaders. The fragment
    shader is updated via the :func:`.gl21.glvector_funcs.updateShaderState`
    function.
    """

    if not self.ready():
        return

    image  = self.image
    shader = self.shader
    opts   = self.opts

    shader.load()

    changed = glvector_funcs.updateShaderState(self)

    # Texture -> value value offsets/scales
    # used by the vertex and fragment shaders
    v1ValXform  = self.v1Texture.voxValXform
    v2ValXform  = self.v2Texture.voxValXform
    v3ValXform  = self.v3Texture.voxValXform
    l1ValXform  = self.l1Texture.voxValXform
    l2ValXform  = self.l2Texture.voxValXform
    l3ValXform  = self.l3Texture.voxValXform

    # Other miscellaneous uniforms
    imageShape    = image.shape[:3]
    resolution    = opts.tensorResolution
    tensorScale   = opts.tensorScale
    xFlip         = opts.orientFlip

    l1           = self.l1
    l1min, l1max = l1.dataRange
    eigValNorm   = 0.5 / max((abs(l1min), abs(l1max)))
    eigValNorm  *= tensorScale / 100.0

    # Define the light position in
    # the eye coordinate system
    lightPos  = np.array([-1, -1, 4], dtype=np.float32)
    lightPos /= np.sqrt(np.sum(lightPos ** 2))

    # Textures used by the vertex shader
    changed |= shader.set('v1Texture', 8)
    changed |= shader.set('v2Texture', 9)
    changed |= shader.set('v3Texture', 10)
    changed |= shader.set('l1Texture', 11)
    changed |= shader.set('l2Texture', 12)
    changed |= shader.set('l3Texture', 13)

    # Texture value -> actual
    # value transformations
    changed |= shader.set('v1ValXform', v1ValXform)
    changed |= shader.set('v2ValXform', v2ValXform)
    changed |= shader.set('v3ValXform', v3ValXform)
    changed |= shader.set('l1ValXform', l1ValXform)
    changed |= shader.set('l2ValXform', l2ValXform)
    changed |= shader.set('l3ValXform', l3ValXform)

    # Other settings
    changed |= shader.set('xFlip',      xFlip)
    changed |= shader.set('imageShape', imageShape)
    changed |= shader.set('eigValNorm', eigValNorm)
    changed |= shader.set('lighting',   opts.lighting)
    changed |= shader.set('lightPos',   lightPos)

    # Vertices of a unit sphere. The vertex
    # shader will transform these vertices
    # into the tensor ellipsoid for each
    # voxel.
    vertices, indices = glroutines.unitSphere(resolution)

    self.nVertices = len(indices)

    shader.setAtt('vertex', vertices)
    shader.setIndices(indices)
    shader.unload()

    return changed


def preDraw(self, xform=None, bbox=None):
    """Must be called before :func:`draw`. Loads the shader programs, and
    does some shader state configuration.
    """

    shader = self.shader
    shader.load()

    # Calculate a transformation matrix for
    # normal vectors - T(I(MV matrix))

    # We transpose mvMat because OpenGL is column-major
    mvMat        = gl.glGetFloatv(gl.GL_MODELVIEW_MATRIX)[:3, :3].T
    v2dMat       = self.opts.getTransform('voxel', 'display')[:3, :3]

    normalMatrix = affine.concat(mvMat, v2dMat)
    normalMatrix = npla.inv(normalMatrix).T

    shader.set('normalMatrix', normalMatrix)

    gl.glEnable(gl.GL_CULL_FACE)
    gl.glEnable(gl.GL_DEPTH_TEST)
    gl.glClear(gl.GL_DEPTH_BUFFER_BIT)
    gl.glCullFace(gl.GL_BACK)


def draw2D(self, zpos, axes, xform=None, bbox=None):
    """Generates voxel coordinates for each tensor to be drawn, does some
    final shader state configuration, and draws the tensors.
    """

    opts   = self.opts
    shader = self.shader
    v2dMat = opts.getTransform('voxel',   'display')

    if xform is None: xform = v2dMat
    else:             xform = affine.concat(v2dMat, xform)

    voxels  = self.generateVoxelCoordinates2D(zpos, axes, bbox)
    nVoxels = len(voxels)

    # Set divisor to 1, so we use one set of
    # voxel coordinates for every sphere drawn
    shader.setAtt('voxel',           voxels, divisor=1)
    shader.set(   'voxToDisplayMat', xform)
    shader.loadAtts()

    arbdi.glDrawElementsInstancedARB(
        gl.GL_QUADS, self.nVertices, gl.GL_UNSIGNED_INT, None, nVoxels)


def draw3D(self, xform=None, bbox=None):
    pass


def postDraw(self, xform=None, bbox=None):
    """Unloads the shader program. """

    self.shader.unloadAtts()
    self.shader.unload()

    gl.glDisable(gl.GL_CULL_FACE)
    gl.glDisable(gl.GL_DEPTH_TEST)



#!/usr/bin/env python
#
# histogramcontrolpanel.py - The HistogramControlPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`HistogramControlPanel` class, a *FSLeyes
control* panel which allows a :class:`.HistogramPanel` to be configured.
"""

import wx

import fsleyes_props                     as props
import fsleyes.tooltips                  as fsltooltips
import fsleyes.strings                   as strings
import fsleyes.plotting.histogramseries  as hseries
import fsleyes.views.histogrampanel      as histogrampanel
from . import                               plotcontrolpanel


class HistogramControlPanel(plotcontrolpanel.PlotControlPanel):
    """The ``HistogramControlPanel`` is a *FSLeyes control* panel which
    allows the user to configure a :class:`.HistogramPanel`. A
    ``HistogramControlPanel`` looks something like the following:

    .. image:: images/histogramcontrolpanel.png
       :scale: 50%
       :align: center
    """

    @staticmethod
    def defaultLayout():
        """Returns a dictionary containing layout settings to be passed to
        :class:`.ViewPanel.togglePanel`.
        """
        return {'location' : wx.RIGHT}


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``HistogramControlPanel`` is only intended to be added to
        :class:`.HistogramPanel` views.
        """
        return [histogrampanel.HistogramPanel]


    def __init__(self, *args, **kwargs):
        """Create a ``HistogramControlPanel``. All arguments are passed
        through to the :meth:`.PlotControlPanel.__init__` method.
        """

        plotcontrolpanel.PlotControlPanel.__init__(self, *args, **kwargs)


    def generateCustomPlotPanelWidgets(self, groupName):
        """Overrides :meth:`.PlotControlPanel.generateCustomPlotPanelWidgets`.
        Adds some widgets to control properties of the
        :class:`.HistogramPanel`.
        """

        hsPanel    = self.plotPanel
        widgetList = self.getWidgetList()
        allWidgets = []
        histProps  = ['histType', 'plotType']

        for prop in histProps:

            kwargs = {}

            kwargs['labels'] = strings.choices[hsPanel, prop]

            widget = props.makeWidget(widgetList, hsPanel, prop, **kwargs)
            allWidgets.append(widget)
            widgetList.AddWidget(
                widget,
                displayName=strings.properties[hsPanel, prop],
                tooltip=fsltooltips.properties[hsPanel, prop],
                groupName=groupName)

        return allWidgets


    def generateCustomDataSeriesWidgets(self, hs, groupName):
        """Overrides :meth:`.PlotControlPanel.generateCustomDataSeriesWidgets`.
        Adds some widgets to control properties of the
        :class:`.HistogramSeries`.
        """

        isimage    = isinstance(hs, hseries.ImageHistogramSeries)
        widgetList = self.getWidgetList()
        allWidgets = []

        autoBin    = props.Widget('autoBin')
        nbins      = props.Widget('nbins',
                                  enabledWhen=lambda i: not i.autoBin,
                                  showLimits=False)
        autoBin    = props.buildGUI(widgetList, hs, autoBin)
        nbins      = props.buildGUI(widgetList, hs, nbins)

        dataRange = props.makeWidget(
            widgetList, hs, 'dataRange',
            labels=[strings.choices['HistogramPanel.dataRange.min'],
                    strings.choices['HistogramPanel.dataRange.max']],
            showLimits=False)

        ignoreZeros     = props.makeWidget(widgetList, hs, 'ignoreZeros')
        includeOutliers = props.makeWidget(widgetList, hs, 'includeOutliers')

        widgetList.AddWidget(ignoreZeros,
                             groupName=groupName,
                             displayName=strings.properties[hs, 'ignoreZeros'],
                             tooltip=fsltooltips.properties[hs, 'ignoreZeros'])
        allWidgets.append(ignoreZeros)

        if isimage:
            showOverlay = props.makeWidget(widgetList, hs, 'showOverlay')
            widgetList.AddWidget(
                showOverlay,
                groupName=groupName,
                displayName=strings.properties[hs, 'showOverlay'],
                tooltip=fsltooltips.properties[hs, 'showOverlay'])
            allWidgets.append(showOverlay)

        widgetList.AddWidget(includeOutliers,
                             groupName=groupName,
                             displayName=strings.properties[hs,
                                                            'includeOutliers'],
                             tooltip=fsltooltips.properties[hs,
                                                            'includeOutliers'])
        widgetList.AddWidget(autoBin,
                             groupName=groupName,
                             displayName=strings.properties[hs, 'autoBin'],
                             tooltip=fsltooltips.properties[hs, 'autoBin'])
        widgetList.AddWidget(nbins,
                             groupName=groupName,
                             displayName=strings.properties[hs, 'nbins'],
                             tooltip=fsltooltips.properties[hs, 'nbins'])
        widgetList.AddWidget(dataRange,
                             groupName=groupName,
                             displayName=strings.properties[hs, 'dataRange'],
                             tooltip=fsltooltips.properties[hs, 'dataRange'])
        allWidgets.append(includeOutliers)
        allWidgets.append(autoBin)
        allWidgets.append(nbins)
        allWidgets.append(dataRange)

        if isinstance(hs, hseries.ComplexHistogramSeries):
            for propName in ['plotReal', 'plotImaginary',
                             'plotMagnitude', 'plotPhase']:
                widget = props.makeWidget(widgetList, hs, propName)
                widgetList.AddWidget(
                    widget,
                    groupName=groupName,
                    displayName=strings.properties[hs, propName],
                    tooltip=fsltooltips.properties[hs, propName])
                allWidgets.append(widget)

        return allWidgets



#!/usr/bin/env python
#
# texture2d.py - The Texture2D class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Texture2D` and :class:`DepthTexture`
classes.
"""


import logging

import numpy     as np
import OpenGL.GL as gl

import fsl.transform.affine as affine
import fsl.data.utils       as dutils
import fsleyes.gl.routines  as glroutines
from . import                  texture


log = logging.getLogger(__name__)


class DepthTexture(texture.Texture):
    """The ``DepthTexture`` class is a 2D ``GL_DEPTH_COMPONENT24`` texture
    which is used by the :class:`.RenderTexture` class.

    A ``DepthTexture`` is configured by setting its :meth:`.Texture.shape`
    property to the desired width/height.
    """


    def __init__(self, name):
        """Create a ``DepthTexture``

        :arg name: Unique name for this texture
        """
        texture.Texture.__init__(self, name, 2, 1, dtype=np.uint32)


    @property
    def dtype(self):
        """Overrides :meth:`.Texture.dtype`. """
        return np.uint32


    @dtype.setter
    def dtype(self):
        """Overrides the :meth:`.Texture.dtype` setter. Raises
        ``NotImplementedError``.
        """
        raise NotImplementedError()


    @property
    def textureType(self):
        """Overrides :meth:`.Texture.textureType`. """
        return gl.GL_UNSIGNED_INT


    @property
    def baseFormat(self):
        """Overrides :meth:`.Texture.baseFormat`. """
        return gl.GL_DEPTH_COMPONENT

    @property
    def internalFormat(self):
        """Overrides :meth:`.Texture.internalFormat`. """
        return gl.GL_DEPTH_COMPONENT24


    @texture.Texture.data.setter
    def data(self, data):
        """Overrides the :meth:`.Texture.data` setter. Raises an error -
        you cannot set data on a ``DepthTexture``.
        """
        raise NotImplementedError('Cannot set data on a DepthTexture')


    def doRefresh(self):
        """Refreshes this ``DepthTexture`` based on the current
        :meth:`.Texture.shape`.
        """

        width, height = self.shape
        ttype         = self.textureType
        intFmt        = self.internalFormat
        baseFmt       = self.baseFormat

        with self.bound():

            gl.glPixelStorei(gl.GL_PACK_ALIGNMENT,   1)
            gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)

            gl.glTexParameteri(gl.GL_TEXTURE_2D,
                               gl.GL_TEXTURE_MAG_FILTER,
                               gl.GL_NEAREST)
            gl.glTexParameteri(gl.GL_TEXTURE_2D,
                               gl.GL_TEXTURE_MIN_FILTER,
                               gl.GL_NEAREST)

            gl.glTexParameteri(gl.GL_TEXTURE_2D,
                               gl.GL_TEXTURE_WRAP_S,
                               gl.GL_CLAMP_TO_EDGE)
            gl.glTexParameteri(gl.GL_TEXTURE_2D,
                               gl.GL_TEXTURE_WRAP_T,
                               gl.GL_CLAMP_TO_EDGE)
            gl.glTexImage2D(gl.GL_TEXTURE_2D,
                            0,
                            intFmt,
                            width,
                            height,
                            0,
                            baseFmt,
                            ttype,
                            None)


class Texture2D(texture.Texture):
    """The ``Texture2D`` class represents a 2D texture. A ``Texture2D``
    instance can be used in one of two ways:

      - Setting the texture data via the :meth:`.Texture.data` method, and then
        drawing it to a scene via :meth:`draw` or :meth:`drawOnBounds`.

      - Setting the texture size via :meth:`.Texture.shape`, and then drawing
        to it by some other means (see e.g. the :class:`.RenderTexture` class,
        a sub-class of ``Texture2D``).
    """


    def __init__(self, name, **kwargs):
        """Create a ``Texture2D`` instance.

        :arg name: Unique name for this ``Texture2D``.
        """

        nvals = kwargs.pop('nvals', 4)

        if nvals not in (1, 3, 4):
            raise ValueError('nvals must be 1, 3 or 4')

        # We keep a copy of the current
        # width/height, so we can detect
        # whether it has changed, and
        # skip unnecessary processing
        self.__width  = None
        self.__height = None

        texture.Texture.__init__(self, name, 2, nvals, **kwargs)


    def doRefresh(self):
        """Overrides :meth:`.Texture.doRefresh`. Configures this ``Texture2D``.
        This includes setting up interpolation, and setting the texture size
        and data.
        """

        data = self.preparedData

        if   data is None:    width, height = self.shape
        elif self.nvals == 1: width, height = data.shape
        else:                 width, height = data.shape[1:]

        if data is not None:
            data = np.array(data.ravel('F'), copy=False)
            data = dutils.makeWriteable(data)

        interp = self.interp

        if interp is None:
            interp = gl.GL_NEAREST

        with self.bound():

            gl.glPixelStorei(gl.GL_PACK_ALIGNMENT,   1)
            gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)

            gl.glTexParameteri(gl.GL_TEXTURE_2D,
                               gl.GL_TEXTURE_MAG_FILTER,
                               interp)
            gl.glTexParameteri(gl.GL_TEXTURE_2D,
                               gl.GL_TEXTURE_MIN_FILTER,
                               interp)

            if self.border is not None:
                gl.glTexParameteri(gl.GL_TEXTURE_2D,
                                   gl.GL_TEXTURE_WRAP_S,
                                   gl.GL_CLAMP_TO_BORDER)
                gl.glTexParameteri(gl.GL_TEXTURE_2D,
                                   gl.GL_TEXTURE_WRAP_T,
                                   gl.GL_CLAMP_TO_BORDER)
                gl.glTexParameterfv(gl.GL_TEXTURE_2D,
                                    gl.GL_TEXTURE_BORDER_COLOR,
                                    np.asarray(self.border, dtype=np.float32))
            else:
                gl.glTexParameteri(gl.GL_TEXTURE_2D,
                                   gl.GL_TEXTURE_WRAP_S,
                                   gl.GL_CLAMP_TO_EDGE)
                gl.glTexParameteri(gl.GL_TEXTURE_2D,
                                   gl.GL_TEXTURE_WRAP_T,
                                   gl.GL_CLAMP_TO_EDGE)

            # If the width and height have not
            # changed, then we don't need to
            # re-define the texture. But we can
            # use glTexSubImage2D if we have
            # data to upload
            if width  == self.__width  and \
               height == self.__height and \
               data is not None:
                gl.glTexSubImage2D(gl.GL_TEXTURE_2D,
                                   0,
                                   0,
                                   0,
                                   width,
                                   height,
                                   self.baseFormat,
                                   self.textureType,
                                   data)

            # If the width and/or height have
            # changed, we need to re-define
            # the texture properties
            else:
                self.__width  = width
                self.__height = height
                gl.glTexImage2D(gl.GL_TEXTURE_2D,
                                0,
                                self.internalFormat,
                                width,
                                height,
                                0,
                                self.baseFormat,
                                self.textureType,
                                data)


    def shapeData(self, data, oldShape=None):
        """Overrides :meth:`.Texture.shapeData`.

        This method is used by ``Texture2D`` sub-classes which are used to
        store 3D image data (e.g. the :class:`.ImageTexture2D` class). It
        shapes the data, ensuring that it is compatible with a 2D texture.

        :arg data:     ``numpy`` array containing the data
        :arg oldShape: Original data shape; if not provided, is taken from
                       ``data``.
        """

        nvals = self.nvals

        # For scalar, 1D or 2D data, we need
        # to make sure the data has a shape
        # compatible with the Texture2D
        if oldShape is None:
            oldShape = data.shape

        if nvals == 1:
            datShape = data.shape
        else:
            oldShape = oldShape[1:]
            datShape = data.shape[1:]

        oldShape = np.array(oldShape)
        datShape = np.array(datShape)

        if   np.all(oldShape         == [1, 1, 1]): newShape = ( 1,  1)
        elif np.all(oldShape[1:]     == [1, 1]):    newShape = (-1,  1)
        elif np.all(oldShape[[0, 2]] == [1, 1]):    newShape = (-1,  1)
        elif np.all(oldShape[:2]     == [1, 1]):    newShape = ( 1, -1)
        elif        oldShape[2]      == 1:          newShape = datShape[:2]
        elif        oldShape[1]      == 1:          newShape = datShape[[0, 2]]
        elif        oldShape[0]      == 1:          newShape = datShape[1:]

        if nvals > 1:
            newShape = [nvals] + list(newShape)

        return data.reshape(newShape)


    def texCoordXform(self, origShape):
        """Overrides :meth:`.Texture.texCoordXform`.

        Returns an affine matrix which encodes a rotation that maps the two
        major axes of the image voxel coordinate system to the first two axes
        of the texture coordinate system.

        This method is used by sub-classes which are being used to store 3D
        image data, e.g. the :class:`.ImageTexture2D` and
        :class:`.SelectionTexture2D` classes.

        If this texture does not have any data yet, this method will return
        ``None``.
        """

        scales  = [1, 1, 1]
        offsets = [0, 0, 0]
        rots    = [0, 0, 0]

        if origShape is None:
            return None

        if self.nvals > 1:
            origShape = origShape[1:]

        # Here we apply a rotation to the
        # coordinates to force the two major
        # voxel axes to map to the first two
        # texture coordinate axes
        if origShape[0] == 1:
            rots      = [0, -np.pi / 2, -np.pi / 2]
        elif origShape[1] == 1:
            rots      = [-np.pi / 2, 0, 0]
            scales[1] = -1

        return affine.compose(scales, offsets, rots)


    def doPatch(self, data, offset):
        """Overrides :meth:`.Texture.doPatch`. Updates part of the texture
        data.
        """
        shape = data.shape
        data  = data.flatten(order='F')

        with self.bound():
            gl.glTexSubImage2D(gl.GL_TEXTURE_2D,
                               0,
                               offset[0],
                               offset[1],
                               shape[0],
                               shape[1],
                               self.baseFormat,
                               self.textureType,
                               data)


    def __prepareCoords(self, vertices, xform=None):
        """Called by :meth:`draw`. Prepares vertices, texture coordinates and
        indices for drawing the texture.

        If ``vertices is None``, it is assumed that the caller has already
        assigned vertices and texture coordinates, either via a shader, or
        via vertex/texture coordinate pointers. In this case,

        :returns: A tuple containing the vertices, texture coordinates, and
                  indices, or ``(None, None, indices)`` if
                  ``vertices is None``
        """

        indices = np.arange(6, dtype=np.uint32)

        if vertices is None:
            return None, None, indices

        if vertices.shape != (6, 3):
            raise ValueError('Six vertices must be provided')

        if xform is not None:
            vertices = affine.transform(vertices, xform)

        vertices  = np.array(vertices, dtype=np.float32).ravel('C')
        texCoords = self.generateTextureCoords()        .ravel('C')

        return vertices, texCoords, indices


    def draw(self,
             vertices=None,
             xform=None,
             textureUnit=None):
        """Draw the contents of this ``Texture2D`` to a region specified by
        the given vertices. The texture is bound to texture unit 0.

        :arg vertices:    A ``numpy`` array of shape ``6 * 3`` specifying the
                          region, made up of two triangles, to which this
                          ``Texture2D`` should be drawn. If ``None``, it is
                          assumed that the vertices and texture coordinates
                          have already been configured (e.g. via a shader
                          program).

        :arg xform:       A transformation to be applied to the vertices.
                          Ignored if ``vertices is None``.

        :arg textureUnit: Texture unit to bind to. Defaults to
                          ``gl.GL_TEXTURE0``.
        """

        if textureUnit is None:
            textureUnit = gl.GL_TEXTURE0

        vertices, texCoords, indices = self.__prepareCoords(vertices, xform)

        with self.bound(textureUnit):
            gl.glClientActiveTexture(textureUnit)
            gl.glTexEnvf(gl.GL_TEXTURE_ENV,
                         gl.GL_TEXTURE_ENV_MODE,
                         gl.GL_REPLACE)

            glfeatures = [gl.GL_TEXTURE_2D, gl.GL_VERTEX_ARRAY]

            # Only enable texture coordinates if we know
            # that there are texture coordinates. Some GL
            # platforms will crash if texcoords are
            # enabled on a texture unit, but no texcoords
            # are loaded.
            if vertices is not None:
                glfeatures.append(gl.GL_TEXTURE_COORD_ARRAY)

            with glroutines.enabled(glfeatures):

                if vertices is not None:
                    gl.glVertexPointer(  3, gl.GL_FLOAT, 0, vertices)
                    gl.glTexCoordPointer(2, gl.GL_FLOAT, 0, texCoords)

                gl.glDrawElements(gl.GL_TRIANGLES, 6, gl.GL_UNSIGNED_INT,
                                  indices)


    def drawOnBounds(self,
                     zpos,
                     xmin,
                     xmax,
                     ymin,
                     ymax,
                     xax,
                     yax,
                     *args,
                     **kwargs):
        """Draws the contents of this ``Texture2D`` to a rectangle.  This is a
        convenience method which creates a set of vertices, and passes them to
        the :meth:`draw` method.

        :arg zpos:  Position along the Z axis, in the display coordinate
                    system.
        :arg xmin:  Minimum X axis coordinate.
        :arg xmax:  Maximum X axis coordinate.
        :arg ymin:  Minimum Y axis coordinate.
        :arg ymax:  Maximum Y axis coordinate.
        :arg xax:   Display space axis which maps to the horizontal screen
                    axis.
        :arg yax:   Display space axis which maps to the vertical screen
                    axis.

        All other arguments are passed to the :meth:`draw` method.
        """

        vertices = self.generateVertices(
            zpos, xmin, xmax, ymin, ymax, xax, yax)
        self.draw(vertices, *args, **kwargs)


    @classmethod
    def generateVertices(
            cls, zpos, xmin, xmax, ymin, ymax, xax, yax, xform=None):
        """Generates a set of vertices suitable for passing to the
        :meth:`.Texture2D.draw` method, for drawing a ``Texture2D`` to a 2D
        canvas.

        :arg zpos:  Position along the Z axis, in the display coordinate
                    system.
        :arg xmin:  Minimum X axis coordinate.
        :arg xmax:  Maximum X axis coordinate.
        :arg ymin:  Minimum Y axis coordinate.
        :arg ymax:  Maximum Y axis coordinate.
        :arg xax:   Display space axis which maps to the horizontal screen
                    axis.
        :arg yax:   Display space axis which maps to the vertical screen
                    axis.
        :arg xform: Transformation matrix to appply to vertices.
        """

        zax              = 3 - xax - yax
        vertices         = np.zeros((6, 3), dtype=np.float32)
        vertices[:, zax] = zpos

        vertices[ 0, [xax, yax]] = [xmin, ymin]
        vertices[ 1, [xax, yax]] = [xmin, ymax]
        vertices[ 2, [xax, yax]] = [xmax, ymin]
        vertices[ 3, [xax, yax]] = [xmax, ymin]
        vertices[ 4, [xax, yax]] = [xmin, ymax]
        vertices[ 5, [xax, yax]] = [xmax, ymax]

        if xform is not None:
            vertices = affine.transform(vertices, xform)

        return vertices


    @classmethod
    def generateTextureCoords(cls):
        """Generates a set of texture coordinates for drawing a
        :class:`Texture2D`. This function is used by the
        :meth:`Texture2D.draw` method.
        """

        texCoords       = np.zeros((6, 2), dtype=np.float32)
        texCoords[0, :] = [0, 0]
        texCoords[1, :] = [0, 1]
        texCoords[2, :] = [1, 0]
        texCoords[3, :] = [1, 0]
        texCoords[4, :] = [0, 1]
        texCoords[5, :] = [1, 1]

        return texCoords


    def getBitmap(self):
        """Returns the data stored in this ``Texture2D`` as a ``numpy.uint8``
        array of shape ``(height, width, 4)``.
        """

        intFmt        = self.baseFormat
        extFmt        = self.textureType
        ndtype        = self.dtype
        nvals         = self.nvals
        width, height = self.shape

        with self.bound():
            data = gl.glGetTexImage(gl.GL_TEXTURE_2D, 0, intFmt, extFmt, None)

        data = np.frombuffer(data, dtype=ndtype)
        data = data.reshape((height, width, nvals))
        data = np.flipud(data)

        return data
